"""
Утилита для управления базой данных PostgreSQL для OPC UA History с поддержкой TimescaleDB.

Этот модуль предоставляет функциональность для:
1. Первоначального создания базы данных
2. Управления учетными данными (с шифрованием)
3. Миграции схемы базы данных к TimescaleDB гипертаблицам
4. Резервного копирования
5. Очистки данных и таблиц
6. Управления TimescaleDB гипертаблицами и автоматическим чанкингом
"""

import asyncio
import json
import logging
import os
import subprocess
import tempfile
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union
from cryptography.fernet import Fernet
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
import base64
import asyncpg
import psycopg
#from psycopg.extensions import ISOLATION_LEVEL_AUTOCOMMIT


class DatabaseManager:
    """
    Менеджер базы данных для OPC UA History с поддержкой TimescaleDB.
    
    Этот класс предоставляет утилиты для управления базой данных PostgreSQL
    вне основного функционала HistoryTimescale.
    
    Использует TimescaleDB гипертаблицы с автоматическим чанкингом по времени
    вместо PostgreSQL партиционирования для лучшей производительности.
    """
    
    def __init__(
        self,
        master_password: str = None,
        config_file: str = "db_config.enc",
        key_file: str = ".db_key",
        encrypted_config: str = None
    ):
        """
        Инициализация менеджера базы данных.
        
        Args:
            master_password: Главный пароль для шифрования/дешифрования
            config_file: Файл для хранения зашифрованной конфигурации
            key_file: Файл для хранения ключа шифрования
            encrypted_config: Зашифрованная конфигурация в виде строки
        """
        self.master_password = master_password
        self.config_file = Path(config_file)
        self.key_file = Path(key_file)
        self.encrypted_config = encrypted_config
        self.logger = logging.getLogger('uapg.db_manager')
        
        # Инициализация шифрования
        if master_password:
            self._init_encryption()
            # Загрузка конфигурации
            self.config = self._load_config()
        elif encrypted_config:
            # Используем зашифрованную конфигурацию
            self.config = self._decrypt_encrypted_config(encrypted_config)
        else:
            self.config = {}
    
    def _init_encryption(self) -> None:
        """Инициализация системы шифрования."""
        if self.key_file.exists():
            # Загружаем существующий ключ
            with open(self.key_file, 'rb') as f:
                self.key = f.read()
        else:
            # Создаем новый ключ на основе master_password
            salt = os.urandom(16)
            kdf = PBKDF2HMAC(
                algorithm=hashes.SHA256(),
                length=32,
                salt=salt,
                iterations=100000,
            )
            key = base64.urlsafe_b64encode(kdf.derive(self.master_password.encode()))
            
            # Сохраняем ключ и соль
            with open(self.key_file, 'wb') as f:
                f.write(salt + key)
            
            self.key = salt + key
        
        self.cipher = Fernet(self.key[16:])  # Используем только ключ, без соли
    
    def _decrypt_encrypted_config(self, encrypted_config: str) -> Dict[str, Any]:
        """Дешифрование конфигурации из строки."""
        try:
            # Декодируем base64
            encrypted_data = base64.b64decode(encrypted_config)
            
            # Создаем временный ключ для расшифровки
            # В реальном приложении здесь должна быть логика получения ключа
            temp_key = Fernet.generate_key()
            temp_cipher = Fernet(temp_key)
            
            # Пытаемся расшифровать
            decrypted_data = temp_cipher.decrypt(encrypted_data)
            return json.loads(decrypted_data.decode())
        except Exception as e:
            self.logger.error(f"Failed to decrypt encrypted config: {e}")
            return {}
    
    def _encrypt_config(self, config: Dict[str, Any]) -> bytes:
        """Шифрование конфигурации."""
        config_str = json.dumps(config, ensure_ascii=False)
        return self.cipher.encrypt(config_str.encode())
    
    def _decrypt_config(self, encrypted_data: bytes) -> Dict[str, Any]:
        """Дешифрование конфигурации."""
        decrypted_data = self.cipher.decrypt(encrypted_data)
        return json.loads(decrypted_data.decode())
    
    def _save_config(self, config: Dict[str, Any]) -> None:
        """Сохранение зашифрованной конфигурации."""
        if hasattr(self, 'cipher'):
            encrypted_data = self._encrypt_config(config)
            with open(self.config_file, 'wb') as f:
                f.write(encrypted_data)
            self.config = config
    
    def _load_config(self) -> Dict[str, Any]:
        """Загрузка конфигурации из файла."""
        if not self.config_file.exists():
            return {}
        
        try:
            with open(self.config_file, 'rb') as f:
                encrypted_data = f.read()
            return self._decrypt_config(encrypted_data)
        except Exception as e:
            self.logger.warning(f"Failed to load config: {e}")
            return {}
    
    def get_connection_params(self) -> Dict[str, Any]:
        """Получение параметров подключения без служебных полей."""
        if not self.config:
            return {}
        
        return {
            'host': self.config['host'],
            'port': self.config['port'],
            'user': self.config['user'],
            'password': self.config['password'],
            'database': self.config['database']
        }

    async def create_database(
        self,
        user: str,
        password: str,
        database: str,
        host: str = "localhost",
        port: int = 5432,
        superuser: str = "postgres",
        superuser_password: str = None,
        enable_timescaledb: bool = True
    ) -> bool:
        """
        Создание базы данных и пользователя с поддержкой TimescaleDB.
        
        Args:
            user: Имя пользователя для создания
            password: Пароль пользователя
            database: Имя базы данных
            host: Хост PostgreSQL
            port: Порт PostgreSQL
            superuser: Суперпользователь для создания БД
            superuser_password: Пароль суперпользователя
            enable_timescaledb: Включить поддержку TimescaleDB
            
        Returns:
            True если успешно создано
        """
        try:
            # Подключение к PostgreSQL как суперпользователь
            if superuser_password:
                conn_params = {
                    'host': host,
                    'port': port,
                    'user': superuser,
                    'password': superuser_password,
                    'database': 'postgres'
                }
            else:
                # Попытка подключения без пароля (для локальной установки)
                conn_params = {
                    'host': host,
                    'port': port,
                    'user': superuser,
                    'database': 'postgres'
                }
            
            # Создание пользователя и базы данных
            conn = psycopg.connect(**conn_params)
            conn.set_isolation_level(psycopg.ISOLATION_LEVEL_AUTOCOMMIT)
            cursor = conn.cursor()
            
            # Создание пользователя
            try:
                cursor.execute(f"CREATE USER {user} WITH PASSWORD '{password}'")
                self.logger.info(f"User {user} created successfully")
            except psycopg.errors.DuplicateObject:
                self.logger.info(f"User {user} already exists")
            
            # Создание базы данных
            try:
                cursor.execute(f"CREATE DATABASE {database} OWNER {user}")
                self.logger.info(f"Database {database} created successfully")
            except psycopg.errors.DuplicateDatabase:
                self.logger.info(f"Database {database} already exists")
            
            cursor.close()
            conn.close()
            
            # Подключение к новой базе данных для настройки
            new_conn_params = {
                'host': host,
                'port': port,
                'user': user,
                'password': password,
                'database': database
            }
            
            # Создание схемы и расширений (только TimescaleDB архитектура)
            await self._setup_timescale_schema(new_conn_params, enable_timescaledb)
            
            # Сохранение конфигурации
            config = {
                'user': user,
                'password': password,
                'database': database,
                'host': host,
                'port': port,
                'created_at': datetime.now().isoformat(),
                'version': '2.0',
                'architecture': 'timescale_hypertables',
                'timescaledb_enabled': enable_timescaledb
            }
            
            if hasattr(self, 'cipher'):
                self._save_config(config)
            
            self.config = config
            
            self.logger.info(f"Database {database} setup completed successfully")
            return True
            
        except Exception as e:
            self.logger.error(f"Failed to create database: {e}")
            return False
    
    async def _setup_timescale_schema(
        self,
        conn_params: Dict[str, Any],
        enable_timescaledb: bool
    ) -> None:
        """Настройка схемы базы данных с TimescaleDB и партиционированными таблицами."""
        try:
            conn = await asyncpg.connect(**conn_params)
            
            # Создание расширений
            if enable_timescaledb:
                try:
                    await conn.execute('CREATE EXTENSION IF NOT EXISTS timescaledb')
                    self.logger.info("✅ TimescaleDB extension enabled")
                except Exception as e:
                    self.logger.warning(f"⚠️  TimescaleDB extension not available: {e}")
                    enable_timescaledb = False
            
            await conn.execute('CREATE EXTENSION IF NOT EXISTS "uuid-ossp"')
            
            # Создание партиционированных таблиц
            await self._create_timescale_tables(conn, enable_timescaledb)
            
            # Создание базовых таблиц метаданных
            await self._create_timescale_metadata_tables(conn)
            
            await conn.close()
            
        except Exception as e:
            self.logger.error(f"Failed to setup TimescaleDB schema: {e}")
            raise
    
    # Удалена поддержка legacy-схемы с множественными таблицами
    
    async def _create_timescale_tables(self, conn: asyncpg.Connection, enable_timescaledb: bool) -> None:
        """Создание гипертаблиц TimescaleDB без PostgreSQL партиционирования."""
        # Таблица для переменных (обычная таблица без PARTITION BY)
        await conn.execute('''
            CREATE TABLE IF NOT EXISTS variables_data (
                variable_id INTEGER NOT NULL,
                time TIMESTAMPTZ NOT NULL,
                value DOUBLE PRECISION NOT NULL,
                quality SMALLINT,
                source_time TIMESTAMPTZ,
                server_time TIMESTAMPTZ DEFAULT NOW()
            )
        ''')
        
        # Таблица для событий (обычная таблица без PARTITION BY)
        await conn.execute('''
            CREATE TABLE IF NOT EXISTS events_data (
                node_id INTEGER NOT NULL,
                event_type_id INTEGER NOT NULL,
                time TIMESTAMPTZ NOT NULL,
                event_data JSONB NOT NULL,
                source_time TIMESTAMPTZ,
                server_time TIMESTAMPTZ DEFAULT NOW()
            )
        ''')
        
        if enable_timescaledb:
            # Создание гипертаблиц TimescaleDB с автоматическим чанкингом
            try:
                # Гипертаблица для переменных
                await conn.execute('''
                    SELECT create_hypertable(
                        'variables_data', 
                        'time',
                        chunk_time_interval => INTERVAL '1 day',
                        if_not_exists => TRUE
                    )
                ''')
                
                # Гипертаблица для событий
                await conn.execute('''
                    SELECT create_hypertable(
                        'events_data', 
                        'time',
                        chunk_time_interval => INTERVAL '1 day',
                        if_not_exists => TRUE
                    )
                ''')
                
                # Включаем columnstore для сжатия переменных
                await conn.execute('''
                    ALTER TABLE variables_data SET (
                        timescaledb.compress,
                        timescaledb.compress_orderby = 'time DESC'
                    )
                ''')
                
                # Включаем columnstore для сжатия событий
                await conn.execute('''
                    ALTER TABLE events_data SET (
                        timescaledb.compress,
                        timescaledb.compress_orderby = 'time DESC'
                    )
                ''')
                
                # Настройка автоматического сжатия
                await conn.execute('''
                    SELECT add_compression_policy(
                        'variables_data', 
                        INTERVAL '7 days',
                        if_not_exists => TRUE
                    )
                ''')
                
                await conn.execute('''
                    SELECT add_compression_policy(
                        'events_data', 
                        INTERVAL '7 days',
                        if_not_exists => TRUE
                    )
                ''')
                
                self.logger.info("✅ Гипертаблицы TimescaleDB созданы с автоматическим чанкингом")
            except Exception as e:
                self.logger.warning(f"⚠️  Не удалось создать гипертаблицы TimescaleDB: {e}")
                self.logger.info("Продолжаем с обычными таблицами PostgreSQL")
        
        # Создание оптимизированных индексов
        await conn.execute('CREATE INDEX IF NOT EXISTS idx_variables_time ON variables_data (time DESC)')
        await conn.execute('CREATE INDEX IF NOT EXISTS idx_variables_variable_time ON variables_data (variable_id, time DESC)')
        await conn.execute('CREATE INDEX IF NOT EXISTS idx_events_time ON events_data (time DESC)')
        await conn.execute('CREATE INDEX IF NOT EXISTS idx_events_node_time ON events_data (node_id, time DESC)')
        await conn.execute('CREATE INDEX IF NOT EXISTS idx_events_event_data ON events_data USING GIN (event_data)')
        
        self.logger.info("✅ Таблицы TimescaleDB созданы с оптимизированными индексами")
    
    async def _create_timescale_metadata_tables(self, conn: asyncpg.Connection) -> None:
        """Создание таблиц метаданных для TimescaleDB архитектуры."""
        # Таблица метаданных переменных
        await conn.execute('''
            CREATE TABLE IF NOT EXISTS variable_metadata (
                id SERIAL PRIMARY KEY,
                variable_id INTEGER NOT NULL UNIQUE,
                node_id TEXT NOT NULL,
                node_name TEXT,
                data_type TEXT,
                created_at TIMESTAMPTZ DEFAULT NOW(),
                retention_period INTERVAL DEFAULT '365 days',
                max_records BIGINT,
                partition_group INTEGER,
                is_active BOOLEAN DEFAULT TRUE
            )
        ''')
        
        # Таблица метаданных типов событий
        await conn.execute('''
            CREATE TABLE IF NOT EXISTS event_type_metadata (
                id SERIAL PRIMARY KEY,
                event_type_id INTEGER NOT NULL,
                event_type_name TEXT,
                source_node_id TEXT NOT NULL,
                fields JSONB NOT NULL,
                created_at TIMESTAMPTZ DEFAULT NOW(),
                retention_period INTERVAL DEFAULT '365 days',
                max_records BIGINT,
                partition_group INTEGER,
                is_active BOOLEAN DEFAULT TRUE,
                UNIQUE(event_type_id, source_node_id)
            )
        ''')
        
        # Таблица версий схемы
        await conn.execute('''
            CREATE TABLE IF NOT EXISTS schema_version (
                id SERIAL PRIMARY KEY,
                version TEXT NOT NULL UNIQUE,
                applied_at TIMESTAMPTZ DEFAULT NOW(),
                description TEXT,
                migration_script TEXT,
                architecture TEXT DEFAULT 'timescale_hypertables'
            )
        ''')
        
        # Создание индексов
        await conn.execute('CREATE INDEX IF NOT EXISTS idx_variable_metadata_variable_id ON variable_metadata(variable_id)')
        await conn.execute('CREATE INDEX IF NOT EXISTS idx_event_type_metadata_source ON event_type_metadata(source_node_id)')
        await conn.execute('CREATE INDEX IF NOT EXISTS idx_event_type_metadata_type ON event_type_metadata(event_type_id)')
        
        # Вставка начальной версии схемы
        await conn.execute('''
            INSERT INTO schema_version (version, description, architecture) 
            VALUES ('2.0', 'TimescaleDB hypertables schema', 'timescale_hypertables') 
            ON CONFLICT (version) DO NOTHING
        ''')
        
        self.logger.info("✅ Таблицы метаданных TimescaleDB созданы")
    
    # Удалена функция создания базовых таблиц для legacy-схемы
    
    async def migrate_to_timescale(self) -> bool:
        """
        Миграция от старой архитектуры к TimescaleDB.
        
        Returns:
            True если миграция успешна
        """
        if not self.config:
            self.logger.error("No database configuration found")
            return False
        
        try:
            conn_params = self.get_connection_params()
            conn = await asyncpg.connect(**conn_params)
            
            # Проверка текущей архитектуры
            current_architecture = await conn.fetchval('''
                SELECT architecture FROM schema_version 
                ORDER BY applied_at DESC LIMIT 1
            ''')
            
            if current_architecture == 'timescale_hypertables':
                self.logger.info("Database already uses TimescaleDB architecture")
                await conn.close()
                return True
            
            self.logger.info("Starting migration to TimescaleDB architecture...")
            
            # Создание новых таблиц
            await self._create_timescale_tables(conn, True)
            await self._create_timescale_metadata_tables(conn)
            
            # Миграция данных (базовая версия)
            await self._migrate_data_to_timescale(conn)
            
            # Обновление версии схемы
            await conn.execute('''
                INSERT INTO schema_version (version, description, architecture) 
                VALUES ('2.0', 'Migrated to TimescaleDB hypertables', 'timescale_hypertables')
                ON CONFLICT (version) DO NOTHING
            ''')
            
            await conn.close()
            
            # Обновление конфигурации
            self.config['version'] = '2.0'
            self.config['architecture'] = 'timescale_hypertables'
            if hasattr(self, 'cipher'):
                self._save_config(self.config)
            
            self.logger.info("✅ Migration to TimescaleDB completed successfully")
            return True
            
        except Exception as e:
            self.logger.error(f"Migration to TimescaleDB failed: {e}")
            return False
    
    async def _migrate_data_to_timescale(self, conn: asyncpg.Connection) -> None:
        """Базовая миграция данных к TimescaleDB архитектуре."""
        try:
            # Получение существующих таблиц переменных
            variable_tables = await conn.fetch('''
                SELECT table_name, node_id FROM variable_metadata 
                WHERE table_name IS NOT NULL
            ''')
            
            for table in variable_tables:
                table_name = table['table_name']
                node_id = table['node_id']
                
                # Проверка существования таблицы
                exists = await conn.fetchval(f'''
                    SELECT 1 FROM information_schema.tables 
                    WHERE table_name = '{table_name}'
                ''')
                
                if exists:
                    # Создание партиции для переменной
                    await self._create_variable_partition(conn, int(node_id) if node_id.isdigit() else 1)
                    
                    # Миграция данных (базовая версия)
                    await self._migrate_variable_data(conn, table_name, int(node_id) if node_id.isdigit() else 1)
            
            self.logger.info("✅ Basic data migration completed")
            
        except Exception as e:
            self.logger.error(f"Data migration failed: {e}")
            raise
    
    async def _create_variable_partition(self, conn: asyncpg.Connection, variable_id: int) -> None:
        """TimescaleDB автоматически управляет чанками - партиции не нужны."""
        # TimescaleDB автоматически создает чанки по времени
        # Дополнительные действия не требуются
        self.logger.info(f"ℹ️  TimescaleDB автоматически управляет чанками для переменной {variable_id}")
        pass
    
    async def _migrate_variable_data(self, conn: asyncpg.Connection, source_table: str, variable_id: int) -> None:
        """Миграция данных переменной из старой таблицы."""
        try:
            # Проверяем структуру исходной таблицы
            columns = await conn.fetch(f'''
                SELECT column_name, data_type 
                FROM information_schema.columns 
                WHERE table_name = '{source_table}'
                ORDER BY ordinal_position
            ''')
            
            # Простая миграция (можно расширить)
            self.logger.info(f"📊 Миграция данных из {source_table} для переменной {variable_id}")
            
            # Здесь должна быть логика миграции данных
            # Для простоты показываем только структуру
            
        except Exception as e:
            self.logger.error(f"Ошибка миграции данных из {source_table}: {e}")
    
    async def backup_database(
        self,
        backup_path: str = None,
        backup_format: str = "custom",
        compression: bool = True
    ) -> Optional[str]:
        """
        Создание резервной копии базы данных.
        
        Args:
            backup_path: Путь для сохранения бэкапа
            backup_format: Формат бэкапа (custom, plain, directory)
            compression: Использовать сжатие
            
        Returns:
            Путь к созданному бэкапу или None при ошибке
        """
        if not self.config:
            self.logger.error("No database configuration found")
            return None
        
        try:
            if not backup_path:
                timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
                backup_path = f"backup_{self.config['database']}_{timestamp}.backup"
            
            # Формирование команды pg_dump
            cmd = [
                'pg_dump',
                f'--host={self.config["host"]}',
                f'--port={self.config["port"]}',
                f'--username={self.config["user"]}',
                f'--dbname={self.config["database"]}',
                f'--format={backup_format}',
                f'--file={backup_path}'
            ]
            
            if compression:
                cmd.append('--compress=9')
            
            # Установка переменной окружения для пароля
            env = os.environ.copy()
            env['PGPASSWORD'] = self.config['password']
            
            # Выполнение команды
            result = subprocess.run(
                cmd,
                env=env,
                capture_output=True,
                text=True
            )
            
            if result.returncode == 0:
                self.logger.info(f"Database backup created: {backup_path}")
                return backup_path
            else:
                self.logger.error(f"Backup failed: {result.stderr}")
                return None
                
        except Exception as e:
            self.logger.error(f"Backup creation failed: {e}")
            return None
    
    async def restore_database(
        self,
        backup_path: str,
        drop_existing: bool = False
    ) -> bool:
        """
        Восстановление базы данных из резервной копии.
        
        Args:
            backup_path: Путь к файлу бэкапа
            drop_existing: Удалить существующую БД перед восстановлением
            
        Returns:
            True если восстановление успешно
        """
        if not self.config:
            self.logger.error("No database configuration found")
            return False
        
        try:
            # Определение формата бэкапа
            if backup_path.endswith('.backup') or backup_path.endswith('.dump'):
                format_flag = '--format=custom'
            elif backup_path.endswith('.sql'):
                format_flag = '--format=plain'
            else:
                format_flag = '--format=custom'
            
            # Формирование команды pg_restore
            cmd = [
                'pg_restore',
                f'--host={self.config["host"]}',
                f'--port={self.config["port"]}',
                f'--username={self.config["user"]}',
                f'--dbname={self.config["database"]}',
                format_flag,
                '--clean',  # Очистка существующих объектов
                '--if-exists',  # Продолжать если объект не существует
                backup_path
            ]
            
            # Установка переменной окружения для пароля
            env = os.environ.copy()
            env['PGPASSWORD'] = self.config['password']
            
            # Выполнение команды
            result = subprocess.run(
                cmd,
                env=env,
                capture_output=True,
                text=True
            )
            
            if result.returncode == 0:
                self.logger.info(f"Database restored from: {backup_path}")
                return True
            else:
                self.logger.error(f"Restore failed: {result.stderr}")
                return False
                
        except Exception as e:
            self.logger.error(f"Database restore failed: {e}")
            return False
    
    async def cleanup_old_data(
        self,
        retention_days: int = 365,
        variable_ids: List[int] = None,
        event_types: List[int] = None
    ) -> bool:
        """
        Очистка старых данных по времени для новой архитектуры.
        
        Args:
            retention_days: Количество дней для хранения данных
            variable_ids: Список ID переменных для очистки (None - все переменные)
            event_types: Список типов событий для очистки (None - все типы)
            
        Returns:
            True если очистка успешна
        """
        if not self.config:
            self.logger.error("No database configuration found")
            return False
        
        try:
            conn_params = self.get_connection_params()
            conn = await asyncpg.connect(**conn_params)
            
            cutoff_date = datetime.now() - timedelta(days=retention_days)
            
            if self.config.get('architecture') == 'timescale_hypertables':
                # Очистка для TimescaleDB архитектуры
                if variable_ids:
                    for variable_id in variable_ids:
                        deleted = await conn.execute('''
                            DELETE FROM variables_data 
                            WHERE variable_id = $1 AND time < $2
                        ''', variable_id, cutoff_date)
                        self.logger.info(f"Cleaned old data for variable {variable_id}")
                else:
                    # Очистка всех данных
                    deleted = await conn.execute('''
                        DELETE FROM variables_data WHERE time < $1
                    ''', cutoff_date)
                    self.logger.info(f"Cleaned all old data from variables_data")
                
                if event_types:
                    for event_type in event_types:
                        deleted = await conn.execute('''
                            DELETE FROM events_data 
                            WHERE event_type_id = $1 AND time < $2
                        ''', event_type, cutoff_date)
                        self.logger.info(f"Cleaned old data for event type {event_type}")
                else:
                    # Очистка всех данных событий
                    deleted = await conn.execute('''
                        DELETE FROM events_data WHERE time < $1
                    ''', cutoff_date)
                    self.logger.info(f"Cleaned all old data from events_data")
            else:
                # Если архитектура не указана, по умолчанию очищаем Timescale таблицы
                if variable_ids:
                    for variable_id in variable_ids:
                        deleted = await conn.execute('''
                            DELETE FROM variables_data 
                            WHERE variable_id = $1 AND time < $2
                        ''', variable_id, cutoff_date)
                        self.logger.info(f"Cleaned old data for variable {variable_id}")
                else:
                    # Очистка всех данных
                    deleted = await conn.execute('''
                        DELETE FROM variables_data WHERE time < $1
                    ''', cutoff_date)
                    self.logger.info(f"Cleaned all old data from variables_data")
                
                if event_types:
                    for event_type in event_types:
                        deleted = await conn.execute('''
                            DELETE FROM events_data 
                            WHERE event_type_id = $1 AND time < $2
                        ''', event_type, cutoff_date)
                        self.logger.info(f"Cleaned old data for event type {event_type}")
                else:
                    # Очистка всех данных событий
                    deleted = await conn.execute('''
                        DELETE FROM events_data WHERE time < $1
                    ''', cutoff_date)
                    self.logger.info(f"Cleaned all old data from events_data")
            
            await conn.close()
            
            self.logger.info(f"Data cleanup completed for records older than {retention_days} days")
            return True
            
        except Exception as e:
            self.logger.error(f"Data cleanup failed: {e}")
            return False
    
    # Удалена очистка данных legacy-схемы
    
    async def remove_node_tables(self, node_ids: List[str]) -> bool:
        """
        Удаление таблиц узлов и связанных метаданных.
        
        Args:
            node_ids: Список ID узлов для удаления
            
        Returns:
            True если удаление успешно
        """
        if not self.config:
            self.logger.error("No database configuration found")
            return False
        
        try:
            conn_params = self.get_connection_params()
            conn = await asyncpg.connect(**conn_params)
            
            if self.config.get('architecture') == 'timescale_hypertables':
                # Для TimescaleDB архитектуры - деактивация переменных
                for node_id in node_ids:
                    await conn.execute('''
                        UPDATE variable_metadata 
                        SET is_active = FALSE 
                        WHERE node_id = $1
                    ''', node_id)
                    self.logger.info(f"Deactivated variable metadata for node {node_id}")
            else:
                # Для старой архитектуры - удаление таблиц
                for node_id in node_ids:
                    # Получение информации о таблице
                    metadata = await conn.fetchrow(
                        'SELECT table_name FROM variable_metadata WHERE node_id = $1',
                        node_id
                    )
                    
                    if metadata:
                        # Удаление таблицы данных
                        await conn.execute(f'DROP TABLE IF EXISTS "{metadata["table_name"]}"')
                        
                        # Удаление метаданных
                        await conn.execute(
                            'DELETE FROM variable_metadata WHERE node_id = $1',
                            node_id
                        )
                        
                        self.logger.info(f"Removed table and metadata for node {node_id}")
            
            await conn.close()
            
            self.logger.info(f"Processed {len(node_ids)} node tables successfully")
            return True
            
        except Exception as e:
            self.logger.error(f"Failed to remove node tables: {e}")
            return False
    
    async def clear_all_data(self) -> bool:
        """
        Полная очистка базы данных от всех данных.
        
        Returns:
            True если очистка успешна
        """
        if not self.config:
            self.logger.error("No database configuration found")
            return False
        
        try:
            conn_params = self.get_connection_params()
            conn = await asyncpg.connect(**conn_params)
            
            if self.config.get('architecture') == 'timescale_hypertables':
                # Очистка для TimescaleDB архитектуры
                await conn.execute('TRUNCATE TABLE variables_data')
                await conn.execute('TRUNCATE TABLE events_data')
                await conn.execute('DELETE FROM variable_metadata')
                await conn.execute('DELETE FROM event_type_metadata')
                
                self.logger.info("Cleared all data from TimescaleDB tables")
            else:
                # Если архитектура не указана, по умолчанию очищаем Timescale таблицы
                await conn.execute('TRUNCATE TABLE variables_data')
                await conn.execute('TRUNCATE TABLE events_data')
                await conn.execute('DELETE FROM variable_metadata')
                await conn.execute('DELETE FROM event_type_metadata')
            
            await conn.close()
            
            self.logger.info("All data cleared from database")
            return True
            
        except Exception as e:
            self.logger.error(f"Failed to clear all data: {e}")
            return False
    
    async def get_database_info(self) -> Dict[str, Any]:
        """
        Получение общей информации о базе данных.
        
        Returns:
            Словарь с информацией о базе данных
        """
        if not self.config:
            return {"error": "No database configuration found"}
        
        try:
            conn_params = self.get_connection_params()
            conn = await asyncpg.connect(**conn_params)
            
            # Проверка версии схемы
            schema_version = await conn.fetchrow('''
                SELECT version, architecture, applied_at, description 
                FROM schema_version 
                ORDER BY applied_at DESC LIMIT 1
            ''')
            
            # Проверка существования основных таблиц
            tables_exist = await conn.fetch('''
                SELECT table_name 
                FROM information_schema.tables 
                WHERE table_schema = 'public' 
                AND table_name IN ('variables_data', 'events_data', 'variable_metadata', 'event_type_metadata')
            ''')
            
            # Количество записей в основных таблицах
            record_counts = {}
            for table in ['variables_data', 'events_data', 'variable_metadata', 'event_type_metadata']:
                try:
                    count = await conn.fetchval(f'SELECT COUNT(*) FROM {table}')
                    record_counts[table] = count
                except Exception:
                    record_counts[table] = "table not found"
            
            await conn.close()
            
            return {
                "schema_version": dict(schema_version) if schema_version else None,
                "tables_exist": [dict(t) for t in tables_exist],
                "record_counts": record_counts,
                "architecture": schema_version['architecture'] if schema_version else "unknown"
            }
            
        except Exception as e:
            self.logger.error(f"Failed to get database info: {e}")
            return {"error": str(e)}
    
    async def get_timescale_info(self) -> Dict[str, Any]:
        """
        Получение информации о TimescaleDB таблицах и партициях.
        
        Returns:
            Словарь с информацией о TimescaleDB
        """
        if not self.config or self.config.get('architecture') != 'timescale_hypertables':
            return {"error": "Database does not use TimescaleDB architecture"}
        
        try:
            conn_params = self.get_connection_params()
            conn = await asyncpg.connect(**conn_params)
            
            # Информация о гипертаблицах
            hypertables = await conn.fetch('''
                SELECT 
                    hypertable_name,
                    num_chunks,
                    compression_enabled,
                    is_compressed
                FROM timescaledb_information.hypertables
            ''')
            
            # Информация о чанках
            chunks = await conn.fetch('''
                SELECT 
                    chunk_name,
                    chunk_schema,
                    range_start,
                    range_end,
                    is_compressed
                FROM timescaledb_information.chunks 
                WHERE hypertable_name IN ('variables_data', 'events_data')
                ORDER BY range_start DESC
                LIMIT 20
            ''')
            
            # Статистика сжатия
            compression_stats = await conn.fetch('''
                SELECT 
                    hypertable_name,
                    compression_status,
                    uncompressed_total_chunks,
                    compressed_total_chunks
                FROM timescaledb_information.compression_settings
            ''')
            
            await conn.close()
            
            return {
                "hypertables": [dict(h) for h in hypertables],
                "chunks": [dict(c) for c in chunks],
                "compression_stats": [dict(cs) for cs in compression_stats]
            }
            
        except Exception as e:
            self.logger.error(f"Failed to get TimescaleDB info: {e}")
            return {"error": str(e)}
    
    async def create_variable_partition(self, variable_id: int) -> bool:
        """
        TimescaleDB автоматически управляет чанками - партиции не нужны.
        
        Args:
            variable_id: ID переменной
            
        Returns:
            True (TimescaleDB автоматически управляет чанками)
        """
        # TimescaleDB автоматически создает чанки по времени
        # Дополнительные действия не требуются
        self.logger.info(f"ℹ️  TimescaleDB автоматически управляет чанками для переменной {variable_id}")
        return True
    
    async def create_event_partition(self, node_id: int) -> bool:
        """
        TimescaleDB автоматически управляет чанками - партиции не нужны.
        
        Args:
            node_id: ID узла
            
        Returns:
            True (TimescaleDB автоматически управляет чанками)
        """
        # TimescaleDB автоматически создает чанки по времени
        # Дополнительные действия не требуются
        self.logger.info(f"ℹ️  TimescaleDB автоматически управляет чанками для узла {node_id}")
        return True
    
    async def enable_compression(self, table_name: str, chunk_time_interval: str = '1 day') -> bool:
        """
        Включение сжатия для TimescaleDB гипертаблицы.
        
        Args:
            table_name: Имя таблицы
            chunk_time_interval: Интервал времени для чанков
            
        Returns:
            True если сжатие включено успешно
        """
        try:
            conn_params = self.get_connection_params()
            conn = await asyncpg.connect(**conn_params)
            
            # Проверяем, является ли таблица гипертаблицей TimescaleDB
            is_hypertable = await conn.fetchval(f'''
                SELECT 1 FROM timescaledb_information.hypertables 
                WHERE hypertable_name = $1
            ''', table_name)
            
            if not is_hypertable:
                self.logger.warning(f"Таблица {table_name} не является гипертаблицей TimescaleDB")
                await conn.close()
                return False
            
            # Установка интервала чанков
            await conn.execute(f'''
                SELECT set_chunk_time_interval('{table_name}', INTERVAL '{chunk_time_interval}')
            ''')
            
            # Включение сжатия
            await conn.execute(f'''
                ALTER TABLE {table_name} SET (
                    timescaledb.compress,
                    timescaledb.compress_orderby = 'time DESC'
                )
            ''')
            
            # Добавление политики сжатия
            await conn.execute(f'''
                SELECT add_compression_policy('{table_name}', INTERVAL '30 days')
            ''')
            
            await conn.close()
            
            self.logger.info(f"✅ Сжатие включено для гипертаблицы {table_name}")
            return True
            
        except Exception as e:
            self.logger.error(f"Failed to enable compression: {e}")
            return False
    
    def change_master_password(self, new_master_password: str) -> bool:
        """
        Изменение главного пароля.
        
        Args:
            new_master_password: Новый главный пароль
            
        Returns:
            True если изменение успешно
        """
        try:
            # Создание нового ключа
            salt = os.urandom(16)
            kdf = PBKDF2HMAC(
                algorithm=hashes.SHA256(),
                length=32,
                salt=salt,
                iterations=100000,
            )
            new_key = base64.urlsafe_b64encode(kdf.derive(new_master_password.encode()))
            
            # Перешифрование конфигурации
            if self.config:
                new_cipher = Fernet(new_key)
                encrypted_data = new_cipher.encrypt(json.dumps(self.config, ensure_ascii=False).encode())
                
                # Сохранение нового ключа
                with open(self.key_file, 'wb') as f:
                    f.write(salt + new_key)
                
                # Сохранение перешифрованной конфигурации
                with open(self.config_file, 'wb') as f:
                    f.write(encrypted_data)
                
                self.key = salt + new_key
                self.cipher = new_cipher
                self.master_password = new_master_password
                
                self.logger.info("Master password changed successfully")
                return True
            
            return False
            
        except Exception as e:
            self.logger.error(f"Failed to change master password: {e}")
            return False
    
    def export_config(self, export_path: str) -> bool:
        """
        Экспорт конфигурации в открытом виде.
        
        Args:
            export_path: Путь для сохранения конфигурации
            
        Returns:
            True если экспорт успешен
        """
        try:
            if self.config:
                with open(export_path, 'w', encoding='utf-8') as f:
                    json.dump(self.config, f, indent=2, ensure_ascii=False)
                
                self.logger.info(f"Configuration exported to {export_path}")
                return True
            
            return False
            
        except Exception as e:
            self.logger.error(f"Failed to export configuration: {e}")
            return False
    
    def import_config(self, import_path: str) -> bool:
        """
        Импорт конфигурации из файла.
        
        Args:
            import_path: Путь к файлу конфигурации
            
        Returns:
            True если импорт успешен
        """
        try:
            with open(import_path, 'r', encoding='utf-8') as f:
                config = json.load(f)
            
            # Шифрование и сохранение
            if hasattr(self, 'cipher'):
                self._save_config(config)
            
            self.config = config
            
            self.logger.info(f"Configuration imported from {import_path}")
            return True
            
        except Exception as e:
            self.logger.error(f"Failed to import configuration: {e}")
            return False


# Утилитарные функции для работы без создания экземпляра класса
async def create_database_standalone(
    user: str,
    password: str,
    database: str,
    host: str = "localhost",
    port: int = 5432,
    superuser: str = "postgres",
    superuser_password: str = None,
    master_password: str = "default_master_password"
) -> bool:
    """
    Создание базы данных без создания экземпляра DatabaseManager.
    
    Args:
        user: Имя пользователя для создания
        password: Пароль пользователя
        database: Имя базы данных
        host: Хост PostgreSQL
        port: Порт PostgreSQL
        superuser: Суперпользователь для создания БД
        superuser_password: Пароль суперпользователя
        master_password: Главный пароль для шифрования
        
    Returns:
        True если успешно создано
    """
    manager = DatabaseManager(master_password)
    return await manager.create_database(
        user, password, database, host, port, superuser, superuser_password
    )


async def backup_database_standalone(
    user: str,
    password: str,
    database: str,
    host: str = "localhost",
    port: int = 5432,
    backup_path: str = None,
    master_password: str = "default_master_password"
) -> Optional[str]:
    """
    Создание резервной копии без создания экземпляра DatabaseManager.
    
    Args:
        user: Имя пользователя
        password: Пароль пользователя
        database: Имя базы данных
        host: Хост PostgreSQL
        port: Порт PostgreSQL
        backup_path: Путь для сохранения бэкапа
        master_password: Главный пароль для шифрования
        
    Returns:
        Путь к созданному бэкапу или None при ошибке
    """
    manager = DatabaseManager(master_password)
    # Установка конфигурации
    manager.config = {
        'user': user,
        'password': password,
        'database': database,
        'host': host,
        'port': port
    }
    return await manager.backup_database(backup_path)


async def migrate_to_timescale_standalone(
    user: str,
    password: str,
    database: str,
    host: str = "localhost",
    port: int = 5432,
    master_password: str = "default_master_password"
) -> bool:
    """
    Миграция к TimescaleDB архитектуре без создания экземпляра DatabaseManager.
    
    Args:
        user: Имя пользователя
        password: Пароль пользователя
        database: Имя базы данных
        host: Хост PostgreSQL
        port: Порт PostgreSQL
        master_password: Главный пароль для шифрования
        
    Returns:
        True если миграция успешна
    """
    manager = DatabaseManager(master_password)
    # Установка конфигурации
    manager.config = {
        'user': user,
        'password': password,
        'database': database,
        'host': host,
        'port': port
    }
    return await manager.migrate_to_timescale()



"""
Тесты для модуля DatabaseManager.
"""

import pytest
import tempfile
import json
from pathlib import Path
from unittest.mock import Mock, patch, AsyncMock

from uapg.db_manager import DatabaseManager


class TestDatabaseManager:
    """Тесты для класса DatabaseManager."""
    
    def setup_method(self):
        """Настройка перед каждым тестом."""
        self.master_password = "test_master_password_123"
        self.temp_dir = tempfile.mkdtemp()
        self.config_file = Path(self.temp_dir) / "db_config.enc"
        self.key_file = Path(self.temp_dir) / ".db_key"
        
        # Создание временного менеджера
        self.db_manager = DatabaseManager(
            self.master_password,
            str(self.config_file),
            str(self.key_file)
        )
    
    def teardown_method(self):
        """Очистка после каждого теста."""
        import shutil
        shutil.rmtree(self.temp_dir, ignore_errors=True)
    
    def test_init_encryption(self):
        """Тест инициализации шифрования."""
        assert self.db_manager.key_file.exists()
        assert self.db_manager.cipher is not None
        assert self.db_manager.master_password == self.master_password
    
    def test_encrypt_decrypt_config(self):
        """Тест шифрования и дешифрования конфигурации."""
        test_config = {
            'user': 'test_user',
            'password': 'test_password',
            'database': 'test_db'
        }
        
        # Шифрование
        encrypted = self.db_manager._encrypt_config(test_config)
        assert isinstance(encrypted, bytes)
        assert encrypted != test_config
        
        # Дешифрование
        decrypted = self.db_manager._decrypt_config(encrypted)
        assert decrypted == test_config
    
    def test_save_load_config(self):
        """Тест сохранения и загрузки конфигурации."""
        test_config = {
            'user': 'test_user',
            'password': 'test_password',
            'database': 'test_db',
            'host': 'localhost',
            'port': 5432
        }
        
        # Сохранение
        self.db_manager._save_config(test_config)
        assert self.config_file.exists()
        
        # Загрузка
        loaded_config = self.db_manager._load_config()
        assert loaded_config == test_config
    
    def test_change_master_password(self):
        """Тест изменения главного пароля."""
        # Сначала сохраняем конфигурацию
        test_config = {'test': 'data'}
        self.db_manager._save_config(test_config)
        
        # Меняем пароль
        new_password = "new_master_password_456"
        success = self.db_manager.change_master_password(new_password)
        
        assert success
        assert self.db_manager.master_password == new_password
        
        # Проверяем, что конфигурация все еще доступна
        loaded_config = self.db_manager._load_config()
        assert loaded_config == test_config
    
    def test_export_import_config(self):
        """Тест экспорта и импорта конфигурации."""
        test_config = {
            'user': 'test_user',
            'password': 'test_password',
            'database': 'test_db'
        }
        
        # Сохраняем конфигурацию
        self.db_manager._save_config(test_config)
        
        # Экспорт
        export_path = Path(self.temp_dir) / "exported_config.json"
        success = self.db_manager.export_config(str(export_path))
        assert success
        assert export_path.exists()
        
        # Проверяем содержимое экспортированного файла
        with open(export_path, 'r', encoding='utf-8') as f:
            exported_data = json.load(f)
        assert exported_data == test_config
        
        # Импорт в новый менеджер
        new_manager = DatabaseManager("new_password")
        success = new_manager.import_config(str(export_path))
        assert success
        assert new_manager.config == test_config
    
    @pytest.mark.asyncio
    async def test_get_database_info_no_config(self):
        """Тест получения информации о БД без конфигурации."""
        info = await self.db_manager.get_database_info()
        assert info == {}
    
    @pytest.mark.asyncio
    async def test_create_database_mock(self):
        """Тест создания базы данных с моком."""
        with patch('uapg.db_manager.psycopg2.connect') as mock_connect, \
             patch('uapg.db_manager.asyncpg.connect') as mock_async_connect:
            
            # Мок для psycopg2
            mock_conn = Mock()
            mock_cursor = Mock()
            mock_conn.cursor.return_value = mock_cursor
            mock_connect.return_value = mock_conn
            
            # Мок для asyncpg
            mock_async_conn = AsyncMock()
            mock_async_connect.return_value = mock_async_conn
            
            success = await self.db_manager.create_database(
                user="test_user",
                password="test_password",
                database="test_db"
            )
            
            assert success
            mock_connect.assert_called_once()
            mock_async_connect.assert_called_once()
    
    @pytest.mark.asyncio
    async def test_backup_database_mock(self):
        """Тест создания резервной копии с моком."""
        # Устанавливаем конфигурацию
        self.db_manager.config = {
            'user': 'test_user',
            'password': 'test_password',
            'database': 'test_db',
            'host': 'localhost',
            'port': 5432
        }
        
        with patch('subprocess.run') as mock_run:
            mock_result = Mock()
            mock_result.returncode = 0
            mock_run.return_value = mock_result
            
            backup_path = await self.db_manager.backup_database()
            
            assert backup_path is not None
            assert "backup_test_db_" in backup_path
            mock_run.assert_called_once()
    
    @pytest.mark.asyncio
    async def test_cleanup_old_data_mock(self):
        """Тест очистки старых данных с моком."""
        # Устанавливаем конфигурацию
        self.db_manager.config = {
            'user': 'test_user',
            'password': 'test_password',
            'database': 'test_db',
            'host': 'localhost',
            'port': 5432
        }
        
        with patch('uapg.db_manager.asyncpg.connect') as mock_connect:
            mock_conn = AsyncMock()
            mock_connect.return_value = mock_conn
            
            # Мок для запросов
            mock_conn.fetch.return_value = []
            mock_conn.fetchrow.return_value = None
            
            success = await self.db_manager.cleanup_old_data(retention_days=30)
            
            assert success
            mock_connect.assert_called_once()
    
    @pytest.mark.asyncio
    async def test_migrate_schema_mock(self):
        """Тест миграции схемы с моком."""
        # Устанавливаем конфигурацию
        self.db_manager.config = {
            'user': 'test_user',
            'password': 'test_password',
            'database': 'test_db',
            'host': 'localhost',
            'port': 5432
        }
        
        with patch('uapg.db_manager.asyncpg.connect') as mock_connect:
            mock_conn = AsyncMock()
            mock_connect.return_value = mock_conn
            
            # Мок для запросов
            mock_conn.fetchval.return_value = "1.0"
            
            migration_scripts = [
                {
                    'version': '1.1',
                    'description': 'Test migration',
                    'sql': 'CREATE INDEX test_idx ON test_table;'
                }
            ]
            
            success = await self.db_manager.migrate_schema('1.1', migration_scripts)
            
            assert success
            mock_connect.assert_called_once()


class TestStandaloneFunctions:
    """Тесты для standalone функций."""
    
    @pytest.mark.asyncio
    async def test_create_database_standalone(self):
        """Тест standalone функции создания БД."""
        from uapg.db_manager import create_database_standalone
        
        with patch('uapg.db_manager.DatabaseManager') as mock_manager_class:
            mock_manager = Mock()
            mock_manager_class.return_value = mock_manager
            mock_manager.create_database = AsyncMock(return_value=True)
            
            success = await create_database_standalone(
                user="test_user",
                password="test_password",
                database="test_db"
            )
            
            assert success
            mock_manager.create_database.assert_called_once()
    
    @pytest.mark.asyncio
    async def test_backup_database_standalone(self):
        """Тест standalone функции создания бэкапа."""
        from uapg.db_manager import backup_database_standalone
        
        with patch('uapg.db_manager.DatabaseManager') as mock_manager_class:
            mock_manager = Mock()
            mock_manager_class.return_value = mock_manager
            mock_manager.backup_database = AsyncMock(return_value="backup.backup")
            
            backup_path = await backup_database_standalone(
                user="test_user",
                password="test_password",
                database="test_db"
            )
            
            assert backup_path == "backup.backup"
            mock_manager.backup_database.assert_called_once()


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



"""
UAPG - OPC UA PostgreSQL History Storage Backend

Модуль для хранения исторических данных OPC UA в PostgreSQL
с поддержкой TimescaleDB для эффективной работы с временными рядами.
"""

__version__ = "0.1.0"
__author__ = "RTS IoT"
__email__ = "panfilov@rts-iot.ru"

from .history_pgsql import HistoryPgSQL
from .history_timescale import HistoryTimescale
from .db_manager import DatabaseManager, create_database_standalone, backup_database_standalone

__all__ = [
    "HistoryPgSQL",
    "HistoryTimescale",
    "DatabaseManager", 
    "create_database_standalone", 
    "backup_database_standalone"
]



import json
import asyncio
import random
import logging
from datetime import datetime, timedelta, timezone
from typing import Any, Iterable, List, Optional, Tuple, Union

import asyncpg
from asyncua import ua
from asyncua.server.history import HistoryStorageInterface
from asyncua.ua.ua_binary import variant_from_binary, variant_to_binary

# Импорт для работы с зашифрованной конфигурацией
from .db_manager import DatabaseManager

# Правильный буфер для побайтного чтения в variant_from_binary
class Buffer:
    def __init__(self, data: bytes) -> None:
        self._data = data
        self._pos = 0
    
    def read(self, n: int) -> bytes:
        chunk = self._data[self._pos:self._pos + n]
        self._pos += n
        return chunk
    
    def copy(self, *args, **kwargs) -> 'Buffer':
        return Buffer(self._data[self._pos:])
    
    def skip(self, n: int) -> None:
        self._pos += n

# Импорт для работы с событиями
from asyncua.common.events import Event

# Импорт фильтрации событий
from .event_filter import apply_event_filter

try:
    from asyncua.server.history import get_event_properties_from_type_node
except ImportError:
    # Fallback для старых версий asyncua
    async def get_event_properties_from_type_node(event_type):
        """Получение свойств события из типа узла"""
        return []

def validate_table_name(name: str) -> None:
    """
    Валидация имени таблицы для предотвращения SQL инъекций.
    
    Args:
        name: Имя таблицы для проверки
        
    Raises:
        ValueError: Если имя таблицы содержит недопустимые символы
    """
    import re
    if not re.match(r'^[\w\-]+$', name):
        raise ValueError(f"Invalid table name: {name}")

class HistoryPgSQL(HistoryStorageInterface):
    """
    Backend для хранения исторических данных OPC UA в PostgreSQL с поддержкой TimescaleDB.
    
    Этот класс реализует интерфейс HistoryStorageInterface и предоставляет
    функциональность для хранения и извлечения исторических данных OPC UA
    в PostgreSQL базе данных с оптимизацией для временных рядов.
    
    Особенности:
    - Таблицы создаются только при инициализации и не изменяются по структуре
    - Отдельные таблицы метаданных для переменных и типов событий
    - Оптимизировано для массовой записи от тысяч источников
    - Использует TimescaleDB для эффективной работы с временными рядами
    
    Attributes:
        max_history_data_response_size (int): Максимальный размер ответа с историческими данными
        logger (logging.Logger): Логгер для записи событий
        _datachanges_period (dict): Словарь периодов хранения данных по узлам
        _conn_params (dict): Параметры подключения к базе данных
        _event_fields (dict): Словарь полей событий по источникам
        _pool (asyncpg.Pool): Пул соединений с базой данных
        _min_size (int): Минимальное количество соединений в пуле
        _max_size (int): Максимальное количество соединений в пуле
        _initialized (bool): Флаг инициализации таблиц метаданных
    """

    def _format_node_id(self, node_id: ua.NodeId) -> str:
        """
        Формирует стандартное имя узла OPC UA в формате ns=X;t=Y.
        
        Args:
            node_id: OPC UA NodeId
            
        Returns:
            str: Строка в формате "ns=X;t=Y"
        """
        # OPC UA использует сокращенные обозначения: i=, s=, g=, b=
        node_id_type_map = {
            ua.NodeIdType.TwoByte: 'i',
            ua.NodeIdType.FourByte: 'i', 
            ua.NodeIdType.Numeric: 'i',
            ua.NodeIdType.String: 's',
            ua.NodeIdType.Guid: 'g',
            ua.NodeIdType.ByteString: 'b'
        }
        node_id_type_char = node_id_type_map.get(node_id.NodeIdType, 'x')
        return f"ns={node_id.NamespaceIndex};{node_id_type_char}={node_id.Identifier}"

    def _get_node_data_type(self, node_id: ua.NodeId, datavalue: Optional[ua.DataValue] = None) -> str:
        """
        Определяет тип данных переменной на основе DataValue или контекста.
        
        Args:
            node_id: OPC UA NodeId переменной
            datavalue: DataValue переменной (опционально)
            
        Returns:
            str: Строковое представление типа данных переменной
        """
        # Если передан DataValue, определяем тип по нему
        if datavalue and hasattr(datavalue, 'Value') and datavalue.Value is not None:
            variant_type = datavalue.Value.VariantType
            if variant_type:
                # Маппинг типов OPC UA VariantType на читаемые названия
                variant_type_map = {
                    ua.VariantType.Boolean: 'Boolean',
                    ua.VariantType.SByte: 'SByte',
                    ua.VariantType.Byte: 'Byte',
                    ua.VariantType.Int16: 'Int16',
                    ua.VariantType.UInt16: 'UInt16',
                    ua.VariantType.Int32: 'Int32',
                    ua.VariantType.UInt32: 'UInt32',
                    ua.VariantType.Int64: 'Int64',
                    ua.VariantType.UInt64: 'UInt64',
                    ua.VariantType.Float: 'Float',
                    ua.VariantType.Double: 'Double',
                    ua.VariantType.String: 'String',
                    ua.VariantType.DateTime: 'DateTime',
                    ua.VariantType.Guid: 'Guid',
                    ua.VariantType.ByteString: 'ByteString',
                    ua.VariantType.XmlElement: 'XmlElement',
                    ua.VariantType.NodeId: 'NodeId',
                    ua.VariantType.ExpandedNodeId: 'ExpandedNodeId',
                    ua.VariantType.StatusCode: 'StatusCode',
                    ua.VariantType.QualifiedName: 'QualifiedName',
                    ua.VariantType.LocalizedText: 'LocalizedText',
                    ua.VariantType.ExtensionObject: 'ExtensionObject',
                    ua.VariantType.DataValue: 'DataValue',
                    ua.VariantType.Variant: 'Variant',
                    ua.VariantType.DiagnosticInfo: 'DiagnosticInfo',
                }
                return variant_type_map.get(variant_type, str(variant_type))
        
        # Если DataValue не передан, пытаемся определить по контексту NodeId
        # Это может быть полезно для предварительной регистрации переменных
        # Например, если знаем, что переменная с определенным NodeId всегда Double
        
        # Маппинг известных переменных по их NodeId
        known_variables = {
            # Пример: если знаем, что переменная с NodeId i=2 всегда Double
            # Можно расширить этот маппинг на основе специфики приложения
        }
        
        # Формируем ключ для поиска
        node_key = f"ns={node_id.NamespaceIndex};i={node_id.Identifier}"
        
        # Возвращаем известный тип или "Unknown" если не определен
        return known_variables.get(node_key, "Unknown")

    def __init__(
        self, 
        user: str = 'postgres', 
        password: str = 'postmaster', 
        database: str = 'opcua', 
        host: str = 'localhost', 
        port: int = 5432,
        min_size: int = 1,
        max_size: int = 10,
        config_file: Optional[str] = None,
        encrypted_config: Optional[str] = None,
        master_password: Optional[str] = None,
        **kwargs
    ) -> None:
        """
        Инициализация HistoryPgSQL.
        
        Args:
            user: Имя пользователя базы данных
            password: Пароль пользователя
            database: Имя базы данных
            host: Хост базы данных
            port: Порт базы данных
            min_size: Минимальное количество соединений в пуле
            max_size: Максимальное количество соединений в пуле
            config_file: Путь к файлу зашифрованной конфигурации
            encrypted_config: Зашифрованная конфигурация в виде строки
            master_password: Главный пароль для расшифровки конфигурации
        """
        self.max_history_data_response_size = 1000
        self.logger = logging.getLogger('uapg.history_pgsql')
        self._datachanges_period = {}
        self._event_fields = {}
        self._pool = None
        self._min_size = min_size
        self._max_size = max_size
        self._initialized = False
        self._pool_lock = asyncio.Lock()
        self._stop_event = asyncio.Event()
        self._reconnect_task = None
        self._reconnect_min_delay = 1.0
        self._reconnect_max_delay = 30.0
        self._was_healthy = True
        self._db_unavailable_since = None
        self._last_throttled_log_at = None
        self._failed_value_saves_counter = 0
        self._failed_event_saves_counter = 0
        
        # Инициализация параметров подключения
        self._conn_params = self._init_connection_params(
            user, password, database, host, port,
            config_file, encrypted_config, master_password, **kwargs
        )
    
    def _init_connection_params(
        self,
        user: str,
        password: str,
        database: str,
        host: str,
        port: int,
        config_file: Optional[str] = None,
        encrypted_config: Optional[str] = None,
        master_password: Optional[str] = None,
        **kwargs
    ) -> dict:
        """
        Инициализация параметров подключения с поддержкой зашифрованной конфигурации.
        
        Args:
            user: Имя пользователя базы данных
            password: Пароль пользователя
            database: Имя базы данных
            host: Хост базы данных
            port: Порт базы данных
            config_file: Путь к файлу зашифрованной конфигурации
            encrypted_config: Зашифрованная конфигурация в виде строки
            master_password: Главный пароль для расшифровки конфигурации
            
        Returns:
            Словарь с параметрами подключения
        """
        # Приоритет: зашифрованная конфигурация > файл конфигурации > прямые параметры
        if encrypted_config and master_password:
            try:
                # Создаем временный DatabaseManager для расшифровки
                temp_manager = DatabaseManager(master_password)
                # Расшифровываем конфигурацию из строки
                decrypted_config = temp_manager._decrypt_config(encrypted_config.encode())
                self.logger.info("Using encrypted configuration from string")
                config_params = {
                    'user': decrypted_config.get('user', user),
                    'password': decrypted_config.get('password', password),
                    'database': decrypted_config.get('database', database),
                    'host': decrypted_config.get('host', host),
                    'port': decrypted_config.get('port', port)
                }
                # Добавляем дополнительные параметры из kwargs
                config_params.update(kwargs)
                return config_params
            except Exception as e:
                self.logger.warning(f"Failed to decrypt configuration string: {e}, using direct parameters")
        
        elif config_file and master_password:
            try:
                # Создаем DatabaseManager для загрузки конфигурации из файла
                temp_manager = DatabaseManager(master_password, config_file)
                if temp_manager.config:
                    self.logger.info(f"Using configuration from file: {config_file}")
                    config_params = {
                        'user': temp_manager.config.get('user', user),
                        'password': temp_manager.config.get('password', password),
                        'database': temp_manager.config.get('database', database),
                        'host': temp_manager.config.get('host', host),
                        'port': temp_manager.config.get('port', port)
                    }
                    # Добавляем дополнительные параметры из kwargs
                    config_params.update(kwargs)
                    return config_params
                else:
                    self.logger.warning(f"Configuration file {config_file} is empty or invalid, using direct parameters")
            except Exception as e:
                self.logger.warning(f"Failed to load configuration from file {config_file}: {e}, using direct parameters")
        
        # Используем прямые параметры как fallback
        self.logger.info("Using direct connection parameters")
        base_params = {
            'user': user,
            'password': password,
            'database': database,
            'host': host,
            'port': port
        }
        # Добавляем дополнительные параметры из kwargs
        base_params.update(kwargs)
        return base_params

    def get_connection_info(self) -> dict:
        """
        Получение информации о текущих параметрах подключения.
        
        Returns:
            Словарь с информацией о подключении
        """
        return {
            'user': self._conn_params['user'],
            'host': self._conn_params['host'],
            'port': self._conn_params['port'],
            'database': self._conn_params['database'],
            'min_size': self._min_size,
            'max_size': self._max_size,
            'initialized': self._initialized
        }

    @classmethod
    def from_config_file(
        cls,
        config_file: str,
        master_password: str,
        min_size: int = 1,
        max_size: int = 10
    ) -> 'HistoryPgSQL':
        """
        Создание экземпляра из файла зашифрованной конфигурации.
        
        Args:
            config_file: Путь к файлу зашифрованной конфигурации
            master_password: Главный пароль для расшифровки
            min_size: Минимальное количество соединений в пуле
            max_size: Максимальное количество соединений в пуле
            
        Returns:
            Экземпляр HistoryPgSQL с загруженной конфигурацией
        """
        return cls(
            config_file=config_file,
            master_password=master_password,
            min_size=min_size,
            max_size=max_size
        )
    
    @classmethod
    def from_encrypted_config(
        cls,
        encrypted_config: str,
        master_password: str,
        min_size: int = 1,
        max_size: int = 10
    ) -> 'HistoryPgSQL':
        """
        Создание экземпляра из зашифрованной конфигурации в виде строки.
        
        Args:
            encrypted_config: Зашифрованная конфигурация в виде строки
            master_password: Главный пароль для расшифровки
            min_size: Минимальное количество соединений в пуле
            max_size: Максимальное количество соединений в пуле
            
        Returns:
            Экземпляр HistoryPgSQL с расшифрованной конфигурацией
        """
        return cls(
            encrypted_config=encrypted_config,
            master_password=master_password,
            min_size=min_size,
            max_size=max_size
        )

    def update_config(
        self,
        config_file: Optional[str] = None,
        encrypted_config: Optional[str] = None,
        master_password: Optional[str] = None
    ) -> bool:
        """
        Обновление конфигурации подключения.
        
        Args:
            config_file: Путь к файлу зашифрованной конфигурации
            encrypted_config: Зашифрованная конфигурация в виде строки
            master_password: Главный пароль для расшифровки конфигурации
            
        Returns:
            True если конфигурация обновлена успешно
        """
        if self._pool:
            self.logger.warning("Cannot update config while pool is active. Call stop() first.")
            return False
        
        try:
            # Сбрасываем флаг инициализации
            self._initialized = False
            
            # Обновляем параметры подключения
            self._conn_params = self._init_connection_params(
                self._conn_params.get('user', 'postgres'),
                self._conn_params.get('password', 'postmaster'),
                self._conn_params.get('database', 'opcua'),
                self._conn_params.get('host', 'localhost'),
                self._conn_params.get('port', 5432),
                config_file, encrypted_config, master_password
            )
            
            self.logger.info("Configuration updated successfully")
            return True
            
        except Exception as e:
            self.logger.error(f"Failed to update configuration: {e}")
            return False

    async def init(self) -> None:
        """Инициализация подключения к базе данных и создание таблиц метаданных."""
        try:
            await self._ensure_pool()

            if not self._initialized:
                await self._create_metadata_tables()
                self._initialized = True

            if self._reconnect_task is None or self._reconnect_task.done():
                self._stop_event.clear()
                self._reconnect_task = asyncio.create_task(self._reconnect_monitor())
                self.logger.info("Reconnect monitor started")

            self.logger.info("HistoryPgSQL initialized successfully")
        except Exception as e:
            self.logger.error(f"Failed to initialize HistoryPgSQL: {e}")
            raise

    def _build_pool_params(self) -> dict:
        pool_params = {
            'user': self._conn_params['user'],
            'password': self._conn_params['password'],
            'database': self._conn_params['database'],
            'host': self._conn_params['host'],
            'port': self._conn_params['port'],
            'min_size': self._min_size,
            'max_size': self._max_size
        }

        exclude_params = {'user', 'password', 'database', 'host', 'port', 'min_size', 'max_size', 'sslmode'}
        for key, value in self._conn_params.items():
            if key not in exclude_params:
                pool_params[key] = value

        if self._conn_params.get('sslmode') == 'disable':
            pool_params['ssl'] = False
        elif self._conn_params.get('sslmode') in ('require', 'verify-ca', 'verify-full'):
            pool_params['ssl'] = True

        return pool_params

    async def _ensure_pool(self) -> None:
        if self._pool and not self._pool._closed:
            return
        async with self._pool_lock:
            if self._pool and not self._pool._closed:
                return
            pool_params = self._build_pool_params()
            self._pool = await asyncpg.create_pool(**pool_params)
            self.logger.info("Connection pool created")

    async def _is_pool_healthy(self) -> bool:
        try:
            await self._ensure_pool()
            async with self._pool.acquire() as conn:
                val = await conn.fetchval('SELECT 1')
                return val == 1
        except Exception:
            return False

    async def _reconnect_monitor(self) -> None:
        delay = self._reconnect_min_delay
        while not self._stop_event.is_set():
            healthy = await self._is_pool_healthy()
            if healthy:
                if not self._was_healthy:
                    self.logger.info("Database connection restored")
                    self._was_healthy = True
                    self._reset_outage_stats()
                delay = self._reconnect_min_delay
                try:
                    await asyncio.wait_for(self._stop_event.wait(), timeout=5.0)
                except asyncio.TimeoutError:
                    pass
                continue

            if self._was_healthy:
                self.logger.error("Database connection lost. The database became unreachable.")
            else:
                self.logger.warning("Database connection unhealthy. Attempting to reconnect...")
            self._was_healthy = False
            try:
                if self._pool:
                    try:
                        await self._pool.close()
                    except Exception:
                        pass
                    self._pool = None
                await self._ensure_pool()
                self.logger.info("Reconnected to database successfully")
                delay = self._reconnect_min_delay
            except Exception as e:
                self.logger.error(f"Reconnect attempt failed: {e}")
                jitter = random.uniform(0, 0.3 * delay)
                await asyncio.sleep(delay + jitter)
                delay = min(delay * 2, self._reconnect_max_delay)

    def _reset_outage_stats(self) -> None:
        if self._failed_value_saves_counter or self._failed_event_saves_counter:
            self.logger.info(
                f"During outage suppressed failures: values={self._failed_value_saves_counter}, events={self._failed_event_saves_counter}"
            )
        self._db_unavailable_since = None
        self._last_throttled_log_at = None
        self._failed_value_saves_counter = 0
        self._failed_event_saves_counter = 0

    def _log_save_failure_throttled(self, kind: str, node_repr: str, error: Exception, datavalue_repr: str = None) -> None:
        now = datetime.now(timezone.utc)
        if self._db_unavailable_since is None:
            self._db_unavailable_since = now
        if kind == 'value':
            self._failed_value_saves_counter += 1
            count = self._failed_value_saves_counter
        else:
            self._failed_event_saves_counter += 1
            count = self._failed_event_saves_counter

        elapsed = now - self._db_unavailable_since
        if elapsed < timedelta(minutes=10):
            # Полная детализация в первые 10 минут
            if datavalue_repr is not None:
                self.logger.error(f"Failed to save {kind} for {node_repr}: {error} \n {datavalue_repr}")
            else:
                self.logger.error(f"Failed to save {kind} for {node_repr}: {error}")
            return

        # После 10 минут — не чаще 1 раза в 10 секунд, с агрегацией
        if self._last_throttled_log_at is None or (now - self._last_throttled_log_at) >= timedelta(seconds=10):
            self._last_throttled_log_at = now
            self.logger.error(
                f"Database still unavailable. Aggregated {kind} save failures: {count}. Latest error: {error}"
            )
            # Сбрасываем только соответствующий счётчик, чтобы считать новый интервал
            if kind == 'value':
                self._failed_value_saves_counter = 0
            else:
                self._failed_event_saves_counter = 0

    async def stop(self) -> None:
        """Остановка и закрытие пула соединений."""
        self._stop_event.set()
        if self._reconnect_task and not self._reconnect_task.done():
            try:
                await self._reconnect_task
            except Exception:
                pass
        self._reconnect_task = None
        if self._pool:
            try:
                await self._pool.close()
            finally:
                self._pool = None
        self.logger.info("HistoryPgSQL stopped")

    async def _execute(self, query: str, *args) -> Any:
        """
        Выполнение SQL запроса.
        
        Args:
            query: SQL запрос
            *args: Аргументы для запроса
            
        Returns:
            Результат выполнения запроса
        """
        await self._ensure_pool()
        try:
            async with self._pool.acquire() as conn:
                return await conn.execute(query, *args)
        except Exception as e:
            self.logger.warning(f"Execute failed, will try to reconnect and retry: {e}")
            await self._force_reconnect()
            async with self._pool.acquire() as conn:
                return await conn.execute(query, *args)

    async def _fetch(self, query: str, *args) -> List[asyncpg.Record]:
        """
        Выполнение SQL запроса с возвратом результатов.
        
        Args:
            query: SQL запрос
            *args: Аргументы для запроса
            
        Returns:
            Список записей
        """
        await self._ensure_pool()
        try:
            async with self._pool.acquire() as conn:
                return await conn.fetch(query, *args)
        except Exception as e:
            self.logger.warning(f"Fetch failed, will try to reconnect and retry: {e}")
            await self._force_reconnect()
            async with self._pool.acquire() as conn:
                return await conn.fetch(query, *args)

    async def _fetchval(self, query: str, *args) -> Any:
        """
        Выполнение SQL запроса с возвратом одного значения.
        
        Args:
            query: SQL запрос
            *args: Аргументы для запроса
            
        Returns:
            Значение
        """
        await self._ensure_pool()
        try:
            async with self._pool.acquire() as conn:
                return await conn.fetchval(query, *args)
        except Exception as e:
            self.logger.warning(f"Fetchval failed, will try to reconnect and retry: {e}")
            await self._force_reconnect()
            async with self._pool.acquire() as conn:
                return await conn.fetchval(query, *args)

    async def _fetchrow(self, query: str, *args) -> Optional[asyncpg.Record]:
        """
        Выполнение SQL запроса с возвратом одной строки.
        
        Args:
            query: SQL запрос
            *args: Аргументы для запроса
            
        Returns:
            Одна строка из результата запроса или None
        """
        await self._ensure_pool()
        try:
            async with self._pool.acquire() as conn:
                return await conn.fetchrow(query, *args)
        except Exception as e:
            self.logger.warning(f"Fetchrow failed, will try to reconnect and retry: {e}")
            await self._force_reconnect()
            async with self._pool.acquire() as conn:
                return await conn.fetchrow(query, *args)

    async def _force_reconnect(self) -> None:
        async with self._pool_lock:
            try:
                if self._pool:
                    try:
                        await self._pool.close()
                    except Exception:
                        pass
                self._pool = None
                await self._ensure_pool()
            except Exception as e:
                self.logger.error(f"Force reconnect failed: {e}")
                raise

    async def _create_metadata_tables(self) -> None:
        """Создание таблиц метаданных для переменных и событий."""
        try:
                        # Таблица метаданных переменных
            await self._execute('''
                CREATE TABLE IF NOT EXISTS variable_metadata (
                    id BIGSERIAL PRIMARY KEY,
                    variable_id BIGINT GENERATED ALWAYS AS (id) STORED,
                    node_id TEXT NOT NULL,
                    data_type TEXT,
                    table_name TEXT NOT NULL,
                    created_at TIMESTAMPTZ DEFAULT NOW(),
                    updated_at TIMESTAMPTZ DEFAULT NOW(),
                    retention_period INTERVAL,
                    max_records INTEGER,
                    UNIQUE(variable_id)
                )
            ''')

            # Таблица метаданных типов событий
            await self._execute('''
                CREATE TABLE IF NOT EXISTS event_type_metadata (
                    id BIGSERIAL PRIMARY KEY,
                    event_type_id BIGINT GENERATED ALWAYS AS (id) STORED,
                    event_type_name TEXT,
                    source_id BIGINT,
                    table_name TEXT NOT NULL,
                    fields JSONB NOT NULL,
                    created_at TIMESTAMPTZ DEFAULT NOW(),
                    updated_at TIMESTAMPTZ DEFAULT NOW(),
                    retention_period INTERVAL,
                    max_records INTEGER,
                    UNIQUE(event_type_id, source_id)
                )
            ''')
            
            # Центральная таблица кэша последних значений переменных
            await self._execute('''
                CREATE TABLE IF NOT EXISTS variable_last_value (
                    node_id TEXT PRIMARY KEY,
                    sourcetimestamp TIMESTAMPTZ NOT NULL,
                    servertimestamp TIMESTAMPTZ NOT NULL,
                    statuscode INTEGER NOT NULL,
                    varianttype INTEGER NOT NULL,
                    variantbinary BYTEA NOT NULL,
                    updated_at TIMESTAMPTZ DEFAULT NOW()
                )
            ''')
            
            # Создаем индексы для метаданных и связей
            # Индексы для bigint полей (оптимизация)
            await self._execute('CREATE INDEX IF NOT EXISTS idx_variable_metadata_variable_id ON variable_metadata(variable_id)')
            await self._execute('CREATE INDEX IF NOT EXISTS idx_event_type_metadata_event_type_id ON event_type_metadata(event_type_id)')
            await self._execute('CREATE INDEX IF NOT EXISTS idx_event_type_metadata_source_id ON event_type_metadata(source_id)')

            # Индексы для текстовых полей
            await self._execute('CREATE INDEX IF NOT EXISTS idx_variable_metadata_table_name ON variable_metadata(table_name)')
            await self._execute('CREATE INDEX IF NOT EXISTS idx_variable_metadata_created ON variable_metadata(created_at)')

            await self._execute('CREATE INDEX IF NOT EXISTS idx_event_type_metadata_table_name ON event_type_metadata(table_name)')
            await self._execute('CREATE INDEX IF NOT EXISTS idx_event_type_metadata_created ON event_type_metadata(created_at)')

            # Составные индексы для оптимизации
            await self._execute('CREATE INDEX IF NOT EXISTS idx_event_type_metadata_type_source ON event_type_metadata(event_type_id, source_id)')
            
            # Уникальные индексы для event_type_metadata
            await self._execute('CREATE UNIQUE INDEX IF NOT EXISTS idx_event_type_metadata_name_null ON event_type_metadata(event_type_name) WHERE source_id IS NULL')
            await self._execute('CREATE UNIQUE INDEX IF NOT EXISTS idx_event_type_metadata_name_source ON event_type_metadata(event_type_name, source_id)')
            
            # Уникальный индекс для variable_metadata по node_id
            await self._execute('CREATE UNIQUE INDEX IF NOT EXISTS idx_variable_metadata_node_id ON variable_metadata(node_id)')
            
            # Индекс для кэш-таблицы последних значений
            await self._execute('CREATE INDEX IF NOT EXISTS idx_variable_last_value_updated ON variable_last_value(updated_at)')

            self.logger.info("Metadata tables created successfully")
        except Exception as e:
            self.logger.error(f"Failed to create metadata tables: {e}")
            raise

    async def _create_variable_table(self, table: str) -> None:
        """
        Создание таблицы переменных.

        Args:
            table: Имя таблицы
        """
        await self._execute(f'''
                CREATE TABLE IF NOT EXISTS "{table}" (
                    _id BIGSERIAL,
                    variable_id BIGINT NOT NULL,  -- для связи с метаданными
                    servertimestamp TIMESTAMPTZ NOT NULL,
                    sourcetimestamp TIMESTAMPTZ NOT NULL,
                    statuscode INTEGER,
                    value TEXT,
                    varianttype INTEGER,
                    variantbinary BYTEA,
                    created_at TIMESTAMPTZ DEFAULT NOW(),
                    PRIMARY KEY (_id, sourcetimestamp)
            )
        ''')

    async def _create_event_table(self, table: str) -> None:
        """
        Создание таблицы событий.

        Args:
            table: Имя таблицы
        """
        await self._execute(f'''
            CREATE TABLE IF NOT EXISTS "{table}" (
                _id BIGSERIAL,
                source_id BIGINT NOT NULL,  -- для связи с метаданными
                event_type_id BIGINT NOT NULL,  -- для связи с метаданными
                _timestamp TIMESTAMPTZ NOT NULL,
                _eventtypename TEXT,
                _eventdata JSONB,
                created_at TIMESTAMPTZ DEFAULT NOW(),
                PRIMARY KEY (_id, _timestamp)
            )
        ''')

    async def _setup_timescale_hypertable(self, table: str, partition_column: str) -> None:
        """
        Настройка TimescaleDB hypertable.
        
        Args:
            table: Имя таблицы
            partition_column: Колонка для партиционирования
        """
        try:
            # Проверяем, доступно ли расширение TimescaleDB
            extension_check = await self._fetchval("SELECT COUNT(*) FROM pg_extension WHERE extname = 'timescaledb'")
            if extension_check == 0:
                self.logger.warning("TimescaleDB extension not found. Creating regular table without hypertable.")
                return
            
            await self._execute(
                f'SELECT create_hypertable(\'{table}\', \'{partition_column}\', if_not_exists => TRUE)'
            )
            self.logger.info(f"TimescaleDB hypertable created for table {table}")
        except Exception as e:
            self.logger.warning(f"Failed to create TimescaleDB hypertable for table {table}: {e}")
            self.logger.info("Continuing with regular table (without TimescaleDB optimization)")

    async def _create_variable_indexes(self, table: str) -> None:
        """
        Создание индексов для таблицы переменных.

        Args:
            table: Имя таблицы
        """
        await self._execute(f'CREATE INDEX IF NOT EXISTS "{table}_id_idx" ON "{table}" (_id)')
        await self._execute(f'CREATE INDEX IF NOT EXISTS "{table}_timestamp_idx" ON "{table}" (sourcetimestamp)')

        # Индексы для bigint полей (оптимизация)
        await self._execute(f'CREATE INDEX IF NOT EXISTS "{table}_variable_id_idx" ON "{table}" (variable_id)')
        await self._execute(f'CREATE INDEX IF NOT EXISTS "{table}_variable_id_timestamp_idx" ON "{table}" (variable_id, sourcetimestamp)')
        
        # Покрывающий индекс для fallback-запросов последнего значения (без variantbinary из-за размера)
        await self._execute(f'CREATE INDEX IF NOT EXISTS "{table}_ts_desc_covering" ON "{table}" (sourcetimestamp DESC) INCLUDE (statuscode, varianttype, servertimestamp)')

    async def _create_event_indexes(self, table: str) -> None:
        """
        Создание индексов для таблицы событий.

        Args:
            table: Имя таблицы
        """
        await self._execute(f'CREATE INDEX IF NOT EXISTS "{table}_id_idx" ON "{table}" (_id)')
        await self._execute(f'CREATE INDEX IF NOT EXISTS "{table}_timestamp_idx" ON "{table}" (_timestamp)')
        await self._execute(f'CREATE INDEX IF NOT EXISTS "{table}_eventtype_idx" ON "{table}" (_eventtypename)')

        # Индексы для bigint полей (оптимизация)
        await self._execute(f'CREATE INDEX IF NOT EXISTS "{table}_source_id_idx" ON "{table}" (source_id)')
        await self._execute(f'CREATE INDEX IF NOT EXISTS "{table}_event_type_id_idx" ON "{table}" (event_type_id)')
        await self._execute(f'CREATE INDEX IF NOT EXISTS "{table}_source_timestamp_idx" ON "{table}" (source_id, _timestamp)')
        await self._execute(f'CREATE INDEX IF NOT EXISTS "{table}_event_type_source_idx" ON "{table}" (event_type_id, source_id)')

    async def _save_variable_metadata(self, node_id: ua.NodeId, table: str, period: Optional[timedelta], count: int) -> int:
        """
        Сохранение метаданных переменной.

        Args:
            node_id: Идентификатор узла
            table: Имя таблицы
            period: Период хранения
            count: Максимальное количество записей

        Returns:
            int: variable_id для использования в таблице истории
        """
        # Сохраняем метаданные переменной
        # Создаем полное имя узла для уникальной идентификации
        node_id_str = self._format_node_id(node_id)
        
        # При регистрации переменной тип данных пока неизвестен
        # Будет обновлен при первом сохранении значения
        data_type = "Unknown"
        
        result = await self._fetchval('''
            INSERT INTO variable_metadata (node_id, data_type, table_name, retention_period, max_records, updated_at)
            VALUES ($1, $2, $3, $4, $5, NOW())
            ON CONFLICT (node_id) DO UPDATE SET
                data_type = EXCLUDED.data_type,
                retention_period = EXCLUDED.retention_period,
                max_records = EXCLUDED.max_records,
                updated_at = NOW()
            RETURNING variable_id
        ''', node_id_str, data_type, table, period, count)

        if result is None:
            # Если не удалось вставить, получаем существующий ID
            node_id_str = self._format_node_id(node_id)
            result = await self._fetchval('''
                SELECT variable_id FROM variable_metadata
                WHERE node_id = $1
                LIMIT 1
            ''', node_id_str)

        return result

    async def _save_event_metadata(self, event_type: ua.NodeId, source_id: ua.NodeId, table: str, fields: List[str], period: Optional[timedelta], count: int) -> Tuple[int, int]:
        """
        Сохранение метаданных события.

        Args:
            event_type: Тип события
            source_id: Идентификатор источника
            table: Имя таблицы
            fields: Список полей
            period: Период хранения
            count: Максимальное количество записей

        Returns:
            Tuple[int, int]: (source_id, event_type_id) для использования в таблице истории
        """
        # Сохраняем метаданные события
        # Сначала создаем запись для типа события без привязки к источнику
        event_type_name = str(getattr(event_type, 'Identifier', str(event_type)))
        
        # Сначала создаем или получаем запись для источника
        source_node_id_str = str(getattr(source_id, 'Identifier', str(source_id)))
        source_db_id = await self._fetchval('''
            INSERT INTO event_type_metadata (event_type_name, table_name, fields, retention_period, max_records, updated_at)
            VALUES ($1, $2, $3, $4, $5, NOW())
            ON CONFLICT (event_type_name) WHERE source_id IS NULL DO UPDATE SET
                fields = EXCLUDED.fields,
                retention_period = EXCLUDED.retention_period,
                max_records = EXCLUDED.max_records,
                updated_at = NOW()
            RETURNING id
        ''', source_node_id_str, table, json.dumps(fields), period, count)
        
        # Теперь создаем запись для типа события с привязкой к источнику
        event_db_id = await self._fetchval('''
            INSERT INTO event_type_metadata (event_type_name, source_id, table_name, fields, retention_period, max_records, updated_at)
            VALUES ($1, $2, $3, $4, $5, $6, NOW())
            ON CONFLICT (event_type_name, source_id) DO UPDATE SET
                fields = EXCLUDED.fields,
                retention_period = EXCLUDED.retention_period,
                max_records = EXCLUDED.max_records,
                updated_at = NOW()
            RETURNING id
        ''', event_type_name, source_db_id, table, json.dumps(fields), period, count)

        return source_db_id, event_db_id

    def _extract_variant_values(self, event_data: dict) -> dict:
        """
        Извлекает значения из Variant объектов для JSON сериализации.
        Преобразует все несериализуемые типы в сериализуемые.
        
        Args:
            event_data: Словарь с данными события, содержащий Variant объекты
            
        Returns:
            Словарь с извлеченными значениями, готовыми для JSON сериализации
        """
        extracted = {}
        for key, value in event_data.items():
            if hasattr(value, 'Value'):
                # Если это Variant, извлекаем значение и рекурсивно обрабатываем
                extracted[key] = self._make_json_serializable(value.Value)
            else:
                # Если не Variant, обрабатываем значение
                extracted[key] = self._make_json_serializable(value)
        return extracted

    def _make_json_serializable(self, value: Any) -> Any:
        """
        Преобразует значение в JSON-сериализуемый тип.
        
        Args:
            value: Значение для преобразования
            
        Returns:
            JSON-сериализуемое значение
        """
        if value is None:
            return None
        elif isinstance(value, (str, int, float, bool)):
            # Базовые типы уже сериализуемы
            return value
        elif isinstance(value, bytes):
            # Байты преобразуем в base64 строку
            import base64
            return base64.b64encode(value).decode('utf-8')
        elif isinstance(value, (list, tuple)):
            # Списки и кортежи обрабатываем рекурсивно
            return [self._make_json_serializable(item) for item in value]
        elif isinstance(value, dict):
            # Словари обрабатываем рекурсивно
            return {k: self._make_json_serializable(v) for k, v in value.items()}
        elif hasattr(value, '__dict__'):
            # Для объектов с атрибутами пытаемся извлечь основные поля
            try:
                # Пытаемся получить строковое представление
                return str(value)
            except:
                # Если не получается, возвращаем имя типа
                return f"{type(value).__name__}"
        else:
            # Для остальных типов используем строковое представление
            try:
                return str(value)
            except:
                return f"{type(value).__name__}"

    def _event_to_binary_map(self, ev_dict: dict) -> dict:
        import base64
        result = {}
        for key, variant in ev_dict.items():
            try:
                # Диагностика для ExtensionObject
                if hasattr(variant, 'VariantType') and variant.VariantType == ua.VariantType.ExtensionObject:
                    self.logger.debug(f"_event_to_binary_map: Processing ExtensionObject for key '{key}': {variant.Value}")
                    try:
                        binary_data = variant_to_binary(variant)
                        self.logger.debug(f"_event_to_binary_map: variant_to_binary success for '{key}', binary length: {len(binary_data)}")
                        result[key] = f"base64:{base64.b64encode(binary_data).decode('utf-8')}"
                    except Exception as e:
                        self.logger.error(f"_event_to_binary_map: variant_to_binary failed for '{key}': {e}")
                        result[key] = None
                else:
                    # Обычная обработка для не-ExtensionObject
                    binary_data = variant_to_binary(variant)
                    result[key] = f"base64:{base64.b64encode(binary_data).decode('utf-8')}"
            except Exception as e:
                self.logger.error(f"_event_to_binary_map: Failed to process key '{key}' with value {variant}: {e}")
                # На всякий случай, если вдруг попадётся не Variant
                try:
                    binary_data = variant_to_binary(ua.Variant(variant))
                    result[key] = f"base64:{base64.b64encode(binary_data).decode('utf-8')}"
                except Exception as e2:
                    self.logger.error(f"_event_to_binary_map: Fallback also failed for '{key}': {e2}")
                    result[key] = None
        return result

    def _binary_map_to_event_values(self, data: dict) -> dict:
        import base64
        result = {}
        for key, b64s in data.items():
            try:
                if b64s is None:
                    self.logger.debug(f"_binary_map_to_event_values: Skipping None value for key '{key}'")
                    result[key] = None
                    continue
                    
                if not isinstance(b64s, str) or not b64s.startswith('base64:'):
                    self.logger.debug(f"_binary_map_to_event_values: Non-base64 value for key '{key}': {type(b64s)} - {b64s}")
                    result[key] = b64s
                    continue
                    
                raw = base64.b64decode(b64s[7:])
                self.logger.debug(f"_binary_map_to_event_values: Decoded binary for key '{key}', length: {len(raw)}")
                
                v = variant_from_binary(Buffer(raw))
                self.logger.debug(f"_binary_map_to_event_values: variant_from_binary success for '{key}': {v}")
                
                # Диагностика для ExtensionObject
                if hasattr(v, 'VariantType') and v.VariantType == ua.VariantType.ExtensionObject:
                    self.logger.debug(f"_binary_map_to_event_values: Recovered ExtensionObject for key '{key}': {v.Value}")
                
                result[key] = v
            except Exception as e:
                self.logger.error(f"_binary_map_to_event_values: Failed to process key '{key}' with value {b64s}: {e}")
                # Фоллбэк: вернуть None
                result[key] = None
        return result

    def _get_table_name(self, node_id: ua.NodeId, table_type: str = "var") -> str:
        """
        Генерация имени таблицы для узла.
        
        Args:
            node_id: Идентификатор узла OPC UA
            table_type: Тип таблицы ("var" для переменных, "evt" для событий)
            
        Returns:
            Имя таблицы в формате "Type_NamespaceIndex_Identifier"
        """
        return f"{table_type}_{node_id.NamespaceIndex}_{node_id.Identifier}"

    async def new_historized_node(
        self, 
        node_id: ua.NodeId, 
        period: Optional[timedelta], 
        count: int = 0
    ) -> None:
        """
        Создание новой таблицы для историзации узла.
        Таблица создается только один раз при инициализации.
        
        Args:
            node_id: Идентификатор узла OPC UA
            period: Период хранения данных (None для бесконечного хранения)
            count: Максимальное количество записей (0 для неограниченного)
        """
        table = self._get_table_name(node_id, "var")
        self.logger.debug(
            "new_historized_node: table=%s node_id=%s period=%s count=%s",
            table, node_id, period, count,
        )
        
        # Сохраняем период хранения
        self._datachanges_period[node_id] = period, count
        
        try:
            validate_table_name(table)
            
            # Создаем таблицу переменных
            await self._create_variable_table(table)
            
            # Преобразуем в hypertable TimescaleDB
            await self._setup_timescale_hypertable(table, 'sourcetimestamp')
            
            # Создаем индексы для производительности
            await self._create_variable_indexes(table)
            
            # Сохраняем метаданные
            await self._save_variable_metadata(node_id, table, period, count)
            
            self.logger.info(f"Variable table {table} created for node {node_id}")
        except Exception as e:
            self.logger.error(f"Failed to create variable table for {node_id}: {e}")
            raise

    async def new_historized_event(
        self, 
        source_id: ua.NodeId, 
        evtypes: List[ua.NodeId], 
        period: Optional[timedelta], 
        count: int = 0
    ) -> None:
        """
        Создание новой таблицы для историзации событий.
        Таблица создается только один раз при инициализации.
        
        Args:
            source_id: Идентификатор источника событий
            evtypes: Список типов событий
            period: Период хранения данных (None для бесконечного хранения)
            count: Максимальное количество записей (0 для неограниченного)
        """
        table = self._get_table_name(source_id, "evt")
        self.logger.debug(
            "new_historized_event: table=%s source_id=%s evtypes=%s period=%s count=%s",
            table, source_id, evtypes, period, count,
        )
        
        # Сохраняем период хранения
        self._datachanges_period[source_id] = period, count
        
        try:
            validate_table_name(table)
            
            # Получаем поля событий
            ev_fields = await self._get_event_fields(evtypes)
            self._event_fields[source_id] = ev_fields
            
            # Создаем таблицу событий с фиксированной структурой
            await self._create_event_table(table)
            
            # Преобразуем в hypertable TimescaleDB
            await self._setup_timescale_hypertable(table, '_timestamp')
            
            # Создаем индексы для производительности
            await self._create_event_indexes(table)
            
            # Сохраняем метаданные для каждого типа события
            for event_type in evtypes:
                await self._save_event_metadata(event_type, source_id, table, ev_fields, period, count)
            
            self.logger.info(f"Event table {table} created for source {source_id}")
        except Exception as e:
            self.logger.error(f"Failed to create event table for {source_id}: {e}")
            raise

    async def _get_event_fields(self, evtypes: List[ua.NodeId]) -> List[str]:
        """
        Получение полей событий из типов узлов.
        
        Args:
            evtypes: Список типов событий
            
        Returns:
            Список имен полей событий
        """
        ev_aggregate_fields = []
        for event_type in evtypes:
            ev_aggregate_fields.extend(await get_event_properties_from_type_node(event_type))
        ev_fields = []
        for field in set(ev_aggregate_fields):
            ev_fields.append((await field.read_display_name()).Text)
        return ev_fields

    async def save_node_value(self, node_id: ua.NodeId, datavalue: ua.DataValue) -> None:
        """
        Сохранение значения узла в историю.
        
        Args:
            node_id: Идентификатор узла OPC UA
            datavalue: Значение данных для сохранения
        """
        table = self._get_table_name(node_id, "var")
        self.logger.debug(
            "save_node_value: table=%s node_id=%s source_ts=%s server_ts=%s status=%s",
            table, node_id, getattr(datavalue, 'SourceTimestamp', None), getattr(datavalue, 'ServerTimestamp', None), getattr(datavalue, 'StatusCode', None),
        )
        
        try:
            validate_table_name(table)
            
            # Простая вставка без проверок структуры
            await self._execute(
                f'INSERT INTO "{table}" (servertimestamp, sourcetimestamp, statuscode, value, varianttype, variantbinary) VALUES ($1, $2, $3, $4, $5, $6)',
                datavalue.ServerTimestamp,
                datavalue.SourceTimestamp,
                datavalue.StatusCode.value,
                str(datavalue.Value.Value),
                int(datavalue.Value.VariantType),
                variant_to_binary(datavalue.Value)
            )
            
            # Обновляем кэш последних значений
            node_id_str = self._format_node_id(node_id)
            await self._execute('''
                INSERT INTO variable_last_value 
                (node_id, sourcetimestamp, servertimestamp, statuscode, varianttype, variantbinary)
                VALUES ($1, $2, $3, $4, $5, $6)
                ON CONFLICT (node_id) DO UPDATE
                    SET sourcetimestamp = EXCLUDED.sourcetimestamp,
                        servertimestamp = EXCLUDED.servertimestamp,
                        statuscode = EXCLUDED.statuscode,
                        varianttype = EXCLUDED.varianttype,
                        variantbinary = EXCLUDED.variantbinary,
                        updated_at = NOW()
                    WHERE variable_last_value.sourcetimestamp <= EXCLUDED.sourcetimestamp
            ''', node_id_str, datavalue.SourceTimestamp, datavalue.ServerTimestamp, 
                datavalue.StatusCode.value, int(datavalue.Value.VariantType), variant_to_binary(datavalue.Value))
            
            # Обновляем тип данных в метаданных на основе реального DataValue только при изменении
            if datavalue and hasattr(datavalue, 'Value') and datavalue.Value is not None:
                actual_data_type = self._get_node_data_type(node_id, datavalue)
                if actual_data_type != "Unknown":
                    # Получаем variable_id для обновления метаданных
                    node_id_str = self._format_node_id(node_id)
                    
                    # Проверяем, изменился ли тип данных
                    current_data_type = await self._fetchval('''
                        SELECT data_type FROM variable_metadata 
                        WHERE node_id = $1
                    ''', node_id_str)
                    
                    # Обновляем только если тип изменился
                    if current_data_type != actual_data_type:
                        await self._execute('''
                            UPDATE variable_metadata 
                            SET data_type = $1, updated_at = NOW() 
                            WHERE node_id = $2
                        ''', actual_data_type, node_id_str)
            
            # Очистка старых данных по периоду хранения
            node_data = self._datachanges_period.get(node_id)
            if node_data is not None and len(node_data) >= 2:
                period, count = node_data[0], node_data[1]
                if period:
                    date_limit = datetime.now(timezone.utc) - period
                    await self._execute(f'DELETE FROM "{table}" WHERE sourcetimestamp < $1', date_limit)
                elif count > 0:
                    # Удаляем лишние записи по количеству
                    await self._execute(f'DELETE FROM "{table}" WHERE _id NOT IN (SELECT _id FROM "{table}" ORDER BY sourcetimestamp DESC LIMIT $1)', count)
                
        except Exception as e:
            # Антиспам логирование при длительной недоступности БД
            self._log_save_failure_throttled('value', str(node_id), e, str(datavalue))

    async def save_event(self, event: Any) -> None:
        """
        Сохранение события в историю.
        
        Args:
            event: Событие OPC UA для сохранения
        """
        self.logger.debug(f"save_event: {type(event)}")
        self.logger.debug(f"save_event: {dir(event)}")
        self.logger.debug(f"save_event: {event.get_event_props_as_fields_dict()}")
        
        if event is None or not hasattr(event, 'SourceNode') or event.SourceNode is None:
            self.logger.error("save_event: invalid event")
            return
        
        table = self._get_table_name(event.SourceNode, "evt")
        event_type = getattr(event, 'EventType', None)
        
        if event_type is None:
            self.logger.error("save_event: event.EventType is None")
            return
        
        try:
            validate_table_name(table)
            
            # Получаем время события
            event_time = getattr(event, 'Time', None) or getattr(event, 'time', None) or datetime.now(timezone.utc)
            
            # Получаем все поля события (Variant) и сериализуем в бинарь (base64)
            raw_event_data = event.get_event_props_as_fields_dict() if hasattr(event, 'get_event_props_as_fields_dict') else {}
            bin_event_data = self._event_to_binary_map(raw_event_data)
            
            # Простая вставка без проверок структуры (JSONB dict, не dumps)
            await self._execute(
                f'INSERT INTO "{table}" (_timestamp, _eventtypename, _eventdata) VALUES ($1, $2, $3)',
                event_time,
                str(event_type),
                json.dumps(bin_event_data)  # asyncpg требует сериализованную строку для JSONB
            )
            
            # Очистка старых данных по периоду хранения
            source_data = self._datachanges_period.get(event.SourceNode)
            if source_data is not None and len(source_data) >= 2:
                period, count = source_data[0], source_data[1]
                if period:
                    date_limit = datetime.now(timezone.utc) - period
                    await self._execute(f'DELETE FROM "{table}" WHERE _timestamp < $1', date_limit)
                elif count > 0:
                    # Удаляем лишние записи по количеству
                    await self._execute(f'DELETE FROM "{table}" WHERE _id NOT IN (SELECT _id FROM "{table}" ORDER BY _timestamp DESC LIMIT $1)', count)
                
        except Exception as e:
            # Антиспам логирование при длительной недоступности БД
            src = getattr(event, 'SourceNode', 'unknown')
            self._log_save_failure_throttled('event', str(src), e)

    async def read_node_history(
        self, 
        node_id: ua.NodeId, 
        start: Optional[datetime], 
        end: Optional[datetime], 
        nb_values: Optional[int], 
        return_bounds: bool = False
    ) -> Tuple[List[ua.DataValue], Optional[datetime]]:
        """
        Чтение истории узла.
        
        Args:
            node_id: Идентификатор узла
            start: Начальное время
            end: Конечное время
            nb_values: Количество значений
            return_bounds: Возвращать ли границы
            
        Returns:
            Кортеж (список значений, время продолжения)
        """
        table = self._get_table_name(node_id, "var")
        start_time, end_time, order, limit = self._get_bounds(start, end, nb_values)
        
        try:
            validate_table_name(table)
            
            # Простой запрос без проверок структуры
            select_sql = f'''
                SELECT servertimestamp, sourcetimestamp, statuscode, value, varianttype, variantbinary
                FROM "{table}" 
                WHERE sourcetimestamp BETWEEN $1 AND $2 
                ORDER BY sourcetimestamp {order} 
                LIMIT $3
            '''
            
            rows = await self._fetch(select_sql, start_time, end_time, limit)
            
            # Преобразуем в DataValue
            results = []
            for row in rows:
                datavalue = ua.DataValue(
                    Value=ua.Variant(row['value'], row['varianttype']),
                    StatusCode=ua.StatusCode(row['statuscode']),
                    SourceTimestamp=row['sourcetimestamp'],
                    ServerTimestamp=row['servertimestamp']
                )
                results.append(datavalue)
            
            # Определяем время продолжения
            cont = None
            if len(results) == limit and len(rows) > 0:
                cont = rows[-1]['sourcetimestamp']
            
            return results, cont
            
        except Exception as e:
            self.logger.error(f"Failed to read node history for {node_id}: {e}")
            return [], None

    async def read_event_history(
        self, 
        source_id: ua.NodeId, 
        start: Optional[datetime], 
        end: Optional[datetime], 
        nb_values: Optional[int], 
        evfilter: Any
    ) -> Tuple[List[Any], Optional[datetime]]:
        """
        Чтение истории событий.
        
        Args:
            source_id: Идентификатор источника событий
            start: Начальное время
            end: Конечное время
            nb_values: Количество значений
            evfilter: Фильтр событий
            
        Returns:
            Кортеж (список событий, время продолжения)
        """
        table = self._get_table_name(source_id, "evt")
        start_time, end_time, order, limit = self._get_bounds(start, end, nb_values)
        
        try:
            validate_table_name(table)
            
            # Простой запрос без проверок структуры
            select_sql = f'''
                SELECT _timestamp, _eventtypename, _eventdata
                FROM "{table}" 
                WHERE _timestamp BETWEEN $1 AND $2 
                ORDER BY _timestamp {order} 
                LIMIT $3
            '''
            
            rows = await self._fetch(select_sql, start_time, end_time, limit)
            
            # Преобразуем в события
            results = []
            for row in rows:
                data = row['_eventdata']
                if isinstance(data, str):
                    data = json.loads(data)
                values = self._binary_map_to_event_values(data)
                #payload = {"Time": row["_timestamp"], "EventType": row["_eventtypename"], **values}
                try:
                    self.logger.debug(f"read_event_history: {values}")
                    event = Event.from_field_dict(values)
                    results.append(event)
                except Exception as e:
                    # Фоллбэк, если from_field_dict недоступен у конкретной реализации Event
                    self.logger.debug(f"read_event_history fallback: {e}")
                    results.append(Event(**values))
            
            # Применяем EventFilter для фильтрации событий
            results = apply_event_filter(results, evfilter)
            
            # Определяем время продолжения
            cont = None
            if len(results) == limit and len(rows) > 0:
                cont = rows[-1]['_timestamp']
            
            return results, cont
        except Exception as e:
            self.logger.error(f"Failed to read event history for {source_id}: {e}")
            return [], None

    @staticmethod
    def _get_bounds(
        start: Optional[datetime], 
        end: Optional[datetime], 
        nb_values: Optional[int]
    ) -> Tuple[datetime, datetime, str, int]:
        """
        Определение границ и параметров для запроса истории.
        
        Args:
            start: Начальное время
            end: Конечное время
            nb_values: Количество значений
            
        Returns:
            Кортеж (начальное время, конечное время, порядок сортировки, лимит)
        """
        order = "ASC"
        if start is None or start == ua.get_win_epoch():
            order = "DESC"
            start = ua.get_win_epoch()
        if end is None or end == ua.get_win_epoch():
            end = datetime.now(timezone.utc) + timedelta(days=1)
        if start < end:
            start_time = start
            end_time = end
        else:
            order = "DESC"
            start_time = end
            end_time = start
        limit = nb_values if nb_values else 10000
        
        return start_time, end_time, order, limit

    async def execute_sql_delete(
        self, 
        condition: str, 
        args: Iterable, 
        table: str, 
        node_id: ua.NodeId
    ) -> None:
        """
        Выполнение SQL запроса удаления данных.
        
        Args:
            condition: SQL условие для удаления
            args: Аргументы для SQL запроса
            table: Имя таблицы
            node_id: Идентификатор узла для логирования
        """
        try:
            validate_table_name(table)
            await self._execute(f'DELETE FROM "{table}" WHERE {condition}', *args)
        except Exception as e:
            self.logger.error(f"Failed to delete data for {node_id}: {e}")

    async def read_last_value(self, node_id: ua.NodeId) -> Optional[ua.DataValue]:
        """
        Быстрое получение последнего сохраненного значения переменной.
        
        Args:
            node_id: Идентификатор узла OPC UA
            
        Returns:
            Последнее значение или None если не найдено
        """
        try:
            node_id_str = self._format_node_id(node_id)
            
            # Сначала пытаемся получить из кэша
            row = await self._fetchrow('''
                SELECT sourcetimestamp, servertimestamp, statuscode, varianttype, variantbinary
                FROM variable_last_value
                WHERE node_id = $1
            ''', node_id_str)
            
            if row is not None:
                # Преобразуем в DataValue
                return ua.DataValue(
                    Value=variant_from_binary(Buffer(row['variantbinary'])),
                    StatusCode_=ua.StatusCode(row['statuscode']),
                    SourceTimestamp=row['sourcetimestamp'],
                    ServerTimestamp=row['servertimestamp']
                )
            
            # Fallback: получаем из таблицы переменных через покрывающий индекс
            table = self._get_table_name(node_id, "var")
            validate_table_name(table)
            
            row = await self._fetchrow(f'''
                SELECT sourcetimestamp, servertimestamp, statuscode, varianttype
                FROM "{table}"
                ORDER BY sourcetimestamp DESC
                LIMIT 1
            ''')
            
            if row is not None:
                # Получаем variantbinary отдельным запросом
                variantbinary_row = await self._fetchrow(f'''
                    SELECT variantbinary
                    FROM "{table}"
                    WHERE sourcetimestamp = $1
                    LIMIT 1
                ''', row['sourcetimestamp'])
                
                if variantbinary_row is not None:
                    # Преобразуем в DataValue
                    return ua.DataValue(
                        Value=variant_from_binary(Buffer(variantbinary_row['variantbinary'])),
                        StatusCode_=ua.StatusCode(row['statuscode']),
                        SourceTimestamp=row['sourcetimestamp'],
                        ServerTimestamp=row['servertimestamp']
                    )
            
            return None
            
        except Exception as e:
            self.logger.error(f"Failed to read last value for {node_id}: {e}")
            return None

    async def read_last_values(self, node_ids: List[ua.NodeId]) -> dict:
        """
        Быстрое получение последних сохраненных значений для списка переменных.
        
        Args:
            node_ids: Список идентификаторов узлов OPC UA
            
        Returns:
            Словарь {node_id: DataValue} или {node_id: None} для отсутствующих
        """
        result = {}
        
        try:
            # Получаем из кэша батчем
            node_id_strs = [self._format_node_id(node_id) for node_id in node_ids]
            
            rows = await self._fetch('''
                SELECT node_id, sourcetimestamp, servertimestamp, statuscode, varianttype, variantbinary
                FROM variable_last_value
                WHERE node_id = ANY($1)
            ''', node_id_strs)
            
            # Обрабатываем результаты из кэша
            cached_node_ids = set()
            for row in rows:
                node_id_str = row['node_id']
                # Находим соответствующий node_id
                for node_id in node_ids:
                    if self._format_node_id(node_id) == node_id_str:
                        cached_node_ids.add(node_id)
                        result[node_id] = ua.DataValue(
                            Value=variant_from_binary(Buffer(row['variantbinary'])),
                            StatusCode_=ua.StatusCode(row['statuscode']),
                            SourceTimestamp=row['sourcetimestamp'],
                            ServerTimestamp=row['servertimestamp']
                        )
                        break
            
            # Fallback для узлов, которых нет в кэше
            missing_node_ids = [node_id for node_id in node_ids if node_id not in cached_node_ids]
            
            for node_id in missing_node_ids:
                try:
                    table = self._get_table_name(node_id, "var")
                    validate_table_name(table)
                    
                    row = await self._fetchrow(f'''
                        SELECT sourcetimestamp, servertimestamp, statuscode, varianttype
                        FROM "{table}"
                        ORDER BY sourcetimestamp DESC
                        LIMIT 1
                    ''')
                    
                    if row is not None:
                        # Получаем variantbinary отдельным запросом
                        variantbinary_row = await self._fetchrow(f'''
                            SELECT variantbinary
                            FROM "{table}"
                            WHERE sourcetimestamp = $1
                            LIMIT 1
                        ''', row['sourcetimestamp'])
                        
                        if variantbinary_row is not None:
                            result[node_id] = ua.DataValue(
                                Value=variant_from_binary(Buffer(variantbinary_row['variantbinary'])),
                                StatusCode_=ua.StatusCode(row['statuscode']),
                                SourceTimestamp=row['sourcetimestamp'],
                                ServerTimestamp=row['servertimestamp']
                            )
                        else:
                            result[node_id] = None
                    else:
                        result[node_id] = None
                        
                except Exception as e:
                    self.logger.warning(f"Failed to read last value for {node_id} from table: {e}")
                    result[node_id] = None
            
            # Заполняем None для узлов, которых вообще нет в истории
            for node_id in node_ids:
                if node_id not in result:
                    result[node_id] = None
            
            return result
            
        except Exception as e:
            self.logger.error(f"Failed to read last values: {e}")
            # Возвращаем None для всех узлов при ошибке
            return {node_id: None for node_id in node_ids}



"""
Тесты для модуля UAPG - OPC UA PostgreSQL History Storage Backend
"""

import pytest
import asyncio
from datetime import datetime, timedelta, timezone
from unittest.mock import AsyncMock, MagicMock, patch

from asyncua import ua
from uapg import HistoryPgSQL


class TestHistoryPgSQL:
    """Тесты для класса HistoryPgSQL."""
    
    @pytest.fixture
    def history(self):
        """Фикстура для создания экземпляра HistoryPgSQL."""
        return HistoryPgSQL(
            user='test_user',
            password='test_password',
            database='test_db',
            host='localhost'
        )
    
    @pytest.fixture
    def mock_connection(self):
        """Фикстура для мок-соединения с базой данных."""
        mock_conn = AsyncMock()
        mock_conn.execute = AsyncMock()
        mock_conn.fetch = AsyncMock()
        mock_conn.close = AsyncMock()
        return mock_conn
    
    @pytest.mark.asyncio
    async def test_init(self, history, mock_connection):
        """Тест инициализации соединения."""
        with patch('asyncpg.connect', return_value=mock_connection):
            await history.init()
            assert history._db == mock_connection
    
    @pytest.mark.asyncio
    async def test_stop(self, history, mock_connection):
        """Тест закрытия соединения."""
        history._db = mock_connection
        await history.stop()
        mock_connection.close.assert_called_once()
    
    def test_get_table_name(self, history):
        """Тест генерации имени таблицы."""
        node_id = ua.NodeId("TestVariable", 1)
        # Тест для переменных (по умолчанию)
        table_name = history._get_table_name(node_id)
        assert table_name == "var_1_TestVariable"
        # Тест для событий
        event_table_name = history._get_table_name(node_id, "evt")
        assert event_table_name == "evt_1_TestVariable"
    
    def test_validate_table_name_valid(self):
        """Тест валидации корректного имени таблицы."""
        from uapg.history_pgsql import validate_table_name
        # Должно пройти без ошибок
        validate_table_name("valid_table_name")
        validate_table_name("table123")
        validate_table_name("table-name")
    
    def test_validate_table_name_invalid(self):
        """Тест валидации некорректного имени таблицы."""
        from uapg.history_pgsql import validate_table_name
        with pytest.raises(ValueError):
            validate_table_name("invalid table name")
        with pytest.raises(ValueError):
            validate_table_name("table;name")
        with pytest.raises(ValueError):
            validate_table_name("table'name")
    
    def test_get_bounds(self, history):
        """Тест определения границ запроса."""
        start = datetime.now(timezone.utc) - timedelta(hours=1)
        end = datetime.now(timezone.utc)
        
        start_time, end_time, order, limit = history._get_bounds(start, end, 100)
        
        assert start_time == start
        assert end_time == end
        assert order == "ASC"
        assert limit == 100
    
    def test_get_bounds_none_values(self, history):
        """Тест определения границ с None значениями."""
        start_time, end_time, order, limit = history._get_bounds(None, None, None)
        
        assert order == "DESC"
        assert limit == 10000
    
    def test_list_to_sql_str(self, history):
        """Тест преобразования списка в SQL строку."""
        test_list = ["column1", "column2", "column3"]
        result = history._list_to_sql_str(test_list)
        expected = '"column1", "column2", "column3"'
        assert result == expected
    
    @pytest.mark.asyncio
    async def test_new_historized_node(self, history, mock_connection):
        """Тест создания таблицы для историзации узла."""
        history._db = mock_connection
        node_id = ua.NodeId(1, "TestVariable")
        
        await history.new_historized_node(node_id, timedelta(days=1), 1000)
        
        # Проверяем, что были вызваны SQL команды
        assert mock_connection.execute.call_count >= 3
    
    @pytest.mark.asyncio
    async def test_save_node_value(self, history, mock_connection):
        """Тест сохранения значения узла."""
        history._db = mock_connection
        node_id = ua.NodeId(1, "TestVariable")
        history._datachanges_period[node_id] = (timedelta(days=1), 1000)
        
        datavalue = ua.DataValue(
            Value=ua.Variant(42.0, ua.VariantType.Double),
            SourceTimestamp=datetime.now(timezone.utc),
            ServerTimestamp=datetime.now(timezone.utc),
            StatusCode_=ua.StatusCode(ua.StatusCodes.Good)
        )
        
        await history.save_node_value(node_id, datavalue)
        
        # Проверяем, что был вызван INSERT
        mock_connection.execute.assert_called()
    
    @pytest.mark.asyncio
    async def test_read_node_history(self, history, mock_connection):
        """Тест чтения истории узла."""
        history._db = mock_connection
        
        # Мокаем результат запроса
        mock_row = {
            'servertimestamp': datetime.now(timezone.utc),
            'sourcetimestamp': datetime.now(timezone.utc),
            'statuscode': 0,
            'variantbinary': b'test_binary_data'
        }
        mock_connection.fetch.return_value = [mock_row]
        
        node_id = ua.NodeId(1, "TestVariable")
        start_time = datetime.now(timezone.utc) - timedelta(hours=1)
        end_time = datetime.now(timezone.utc)
        
        results, continuation = await history.read_node_history(
            node_id, start_time, end_time, 100
        )
        
        assert isinstance(results, list)
        assert continuation is None or isinstance(continuation, datetime)


class TestBuffer:
    """Тесты для класса Buffer."""
    
    def test_buffer_init(self):
        """Тест инициализации буфера."""
        from uapg.history_pgsql import Buffer
        data = b"test_data"
        buffer = Buffer(data)
        assert buffer.data == data
        assert buffer.pos == 0
    
    def test_buffer_read(self):
        """Тест чтения из буфера."""
        from uapg.history_pgsql import Buffer
        data = b"test_data"
        buffer = Buffer(data)
        
        result = buffer.read(4)
        assert result == b"test"
        assert buffer.pos == 4
        
        result = buffer.read(4)
        assert result == b"_dat"
        assert buffer.pos == 8


if __name__ == "__main__":
    pytest.main([__file__]) 


"""
Тесты для пула подключений PostgreSQL в UAPG.
"""

import pytest
import asyncio
from unittest.mock import Mock, patch, AsyncMock
from uapg.history_pgsql import HistoryPgSQL
from asyncua import ua


class TestConnectionPool:
    """Тесты для пула подключений."""
    
    @pytest.fixture
    def history_storage(self):
        """Создание экземпляра HistoryPgSQL для тестирования."""
        return HistoryPgSQL(
            user='test_user',
            password='test_password',
            database='test_db',
            host='test_host',
            port=5432,
            min_size=2,
            max_size=5
        )
    
    @pytest.fixture
    def mock_pool(self):
        """Мок пула подключений."""
        pool = Mock()
        pool.get_min_size.return_value = 2
        pool.get_max_size.return_value = 5
        pool.get_size.return_value = 3
        pool.get_free_size.return_value = 1
        return pool
    
    def test_constructor_parameters(self, history_storage):
        """Тест параметров конструктора."""
        assert history_storage._min_size == 2
        assert history_storage._max_size == 5
        assert history_storage._conn_params['port'] == 5432
        assert history_storage._pool is None
    
    @pytest.mark.asyncio
    async def test_init_creates_pool(self, history_storage):
        """Тест создания пула при инициализации."""
        with patch('asyncpg.create_pool') as mock_create_pool:
            mock_pool = AsyncMock()
            mock_create_pool.return_value = mock_pool
            
            await history_storage.init()
            
            mock_create_pool.assert_called_once_with(
                user='test_user',
                password='test_password',
                database='test_db',
                host='test_host',
                port=5432,
                min_size=2,
                max_size=5,
                command_timeout=60,
                statement_cache_size=0
            )
            assert history_storage._pool == mock_pool
    
    @pytest.mark.asyncio
    async def test_init_handles_errors(self, history_storage):
        """Тест обработки ошибок при инициализации."""
        with patch('asyncpg.create_pool') as mock_create_pool:
            mock_create_pool.side_effect = Exception("Connection failed")
            
            with pytest.raises(Exception, match="Connection failed"):
                await history_storage.init()
            
            assert history_storage._pool is None
    
    @pytest.mark.asyncio
    async def test_stop_closes_pool(self, history_storage, mock_pool):
        """Тест закрытия пула."""
        history_storage._pool = mock_pool
        
        await history_storage.stop()
        
        mock_pool.close.assert_called_once()
    
    @pytest.mark.asyncio
    async def test_stop_handles_none_pool(self, history_storage):
        """Тест закрытия при отсутствии пула."""
        history_storage._pool = None
        
        # Не должно вызывать ошибку
        await history_storage.stop()
    
    @pytest.mark.asyncio
    async def test_execute_uses_pool(self, history_storage, mock_pool):
        """Тест использования пула для выполнения запросов."""
        history_storage._pool = mock_pool
        
        # Мокаем контекстный менеджер
        mock_conn = AsyncMock()
        mock_pool.acquire.return_value.__aenter__.return_value = mock_conn
        mock_pool.acquire.return_value.__aexit__.return_value = None
        
        await history_storage._execute("SELECT 1", "param1")
        
        mock_pool.acquire.assert_called_once()
        mock_conn.execute.assert_called_once_with("SELECT 1", "param1")
    
    @pytest.mark.asyncio
    async def test_fetch_uses_pool(self, history_storage, mock_pool):
        """Тест использования пула для выборки данных."""
        history_storage._pool = mock_pool
        
        mock_conn = AsyncMock()
        mock_pool.acquire.return_value.__aenter__.return_value = mock_conn
        mock_pool.acquire.return_value.__aexit__.return_value = None
        
        await history_storage._fetch("SELECT * FROM table", "param1")
        
        mock_pool.acquire.assert_called_once()
        mock_conn.fetch.assert_called_once_with("SELECT * FROM table", "param1")
    
    @pytest.mark.asyncio
    async def test_fetchval_uses_pool(self, history_storage, mock_pool):
        """Тест использования пула для получения одного значения."""
        history_storage._pool = mock_pool
        
        mock_conn = AsyncMock()
        mock_pool.acquire.return_value.__aenter__.return_value = mock_conn
        mock_pool.acquire.return_value.__aexit__.return_value = None
        
        await history_storage._fetchval("SELECT COUNT(*) FROM table")
        
        mock_pool.acquire.assert_called_once()
        mock_conn.fetchval.assert_called_once_with("SELECT COUNT(*) FROM table")
    
    def test_default_parameters(self):
        """Тест параметров по умолчанию."""
        history = HistoryPgSQL()
        
        assert history._min_size == 5
        assert history._max_size == 20
        assert history._conn_params['port'] == 5432
    
    def test_custom_parameters(self):
        """Тест пользовательских параметров."""
        history = HistoryPgSQL(
            min_size=10,
            max_size=100,
            port=5433
        )
        
        assert history._min_size == 10
        assert history._max_size == 100
        assert history._conn_params['port'] == 5433


class TestPoolIntegration:
    """Тесты интеграции с реальным пулом подключений."""
    
    @pytest.mark.asyncio
    async def test_pool_lifecycle(self):
        """Тест жизненного цикла пула."""
        history = HistoryPgSQL(
            user='test_user',
            password='test_password',
            database='test_db',
            host='test_host',
            min_size=1,
            max_size=3
        )
        
        # Проверяем, что пул не создан до инициализации
        assert history._pool is None
        
        # Мокаем создание пула
        with patch('asyncpg.create_pool') as mock_create_pool:
            mock_pool = AsyncMock()
            mock_create_pool.return_value = mock_pool
            
            await history.init()
            
            # Проверяем, что пул создан
            assert history._pool == mock_pool
            
            # Проверяем, что пул закрывается при остановке
            await history.stop()
            mock_pool.close.assert_called_once()


if __name__ == "__main__":
    pytest.main([__file__])



"""
Тесты для фильтрации событий OPC UA.
"""

import unittest
from asyncua import ua
from asyncua.common.events import Event
from asyncua.common.event_objects import BaseEvent
from uapg.event_filter import (
    EventFilterEvaluator,
    apply_event_filter,
    create_source_node_filter,
    create_event_type_filter,
    create_property_filter,
)


class TestEventFilterBasic(unittest.TestCase):
    """Базовые тесты для фильтрации событий."""
    
    def setUp(self):
        """Подготовка тестовых данных."""
        # Создаем тестовые события
        self.event1 = BaseEvent()
        self.event1.Severity = 300
        self.event1.Message = ua.LocalizedText("Normal operation")
        self.event1.SourceNode = ua.NodeId(1234, 2)
        self.event1.EventType = ua.NodeId(ua.ObjectIds.BaseEventType)
        
        self.event2 = BaseEvent()
        self.event2.Severity = 700
        self.event2.Message = ua.LocalizedText("High severity alarm")
        self.event2.SourceNode = ua.NodeId(1234, 2)
        self.event2.EventType = ua.NodeId(ua.ObjectIds.AlarmConditionType)
        
        self.event3 = BaseEvent()
        self.event3.Severity = 900
        self.event3.Message = ua.LocalizedText("Critical error")
        self.event3.SourceNode = ua.NodeId(5678, 2)
        self.event3.EventType = ua.NodeId(ua.ObjectIds.BaseEventType)
        
        self.events = [self.event1, self.event2, self.event3]
    
    def test_no_filter(self):
        """Тест без фильтра - должны вернуться все события."""
        result = apply_event_filter(self.events, None)
        self.assertEqual(len(result), 3)
    
    def test_empty_filter(self):
        """Тест с пустым фильтром - должны вернуться все события."""
        evfilter = ua.EventFilter()
        result = apply_event_filter(self.events, evfilter)
        self.assertEqual(len(result), 3)
    
    def test_severity_filter_greater_than(self):
        """Тест фильтрации по Severity > 500."""
        evfilter = create_property_filter(
            "Severity",
            ua.FilterOperator.GreaterThan,
            500
        )
        result = apply_event_filter(self.events, evfilter)
        self.assertEqual(len(result), 2)  # event2 и event3
        self.assertIn(self.event2, result)
        self.assertIn(self.event3, result)
    
    def test_severity_filter_greater_than_or_equal(self):
        """Тест фильтрации по Severity >= 700."""
        evfilter = create_property_filter(
            "Severity",
            ua.FilterOperator.GreaterThanOrEqual,
            700
        )
        result = apply_event_filter(self.events, evfilter)
        self.assertEqual(len(result), 2)  # event2 и event3
    
    def test_severity_filter_less_than(self):
        """Тест фильтрации по Severity < 500."""
        evfilter = create_property_filter(
            "Severity",
            ua.FilterOperator.LessThan,
            500
        )
        result = apply_event_filter(self.events, evfilter)
        self.assertEqual(len(result), 1)  # только event1
        self.assertIn(self.event1, result)
    
    def test_severity_filter_equals(self):
        """Тест фильтрации по Severity == 700."""
        evfilter = create_property_filter(
            "Severity",
            ua.FilterOperator.Equals,
            700
        )
        result = apply_event_filter(self.events, evfilter)
        self.assertEqual(len(result), 1)  # только event2
        self.assertIn(self.event2, result)
    
    def test_source_node_filter(self):
        """Тест фильтрации по SourceNode."""
        evfilter = create_source_node_filter(ua.NodeId(1234, 2))
        result = apply_event_filter(self.events, evfilter)
        self.assertEqual(len(result), 2)  # event1 и event2
        self.assertIn(self.event1, result)
        self.assertIn(self.event2, result)
    
    def test_event_type_filter(self):
        """Тест фильтрации по EventType."""
        evfilter = create_event_type_filter([
            ua.NodeId(ua.ObjectIds.AlarmConditionType)
        ])
        result = apply_event_filter(self.events, evfilter)
        self.assertEqual(len(result), 1)  # только event2
        self.assertIn(self.event2, result)
    
    def test_event_type_filter_multiple(self):
        """Тест фильтрации по нескольким типам событий."""
        evfilter = create_event_type_filter([
            ua.NodeId(ua.ObjectIds.BaseEventType),
            ua.NodeId(ua.ObjectIds.AlarmConditionType)
        ])
        result = apply_event_filter(self.events, evfilter)
        # Все события должны пройти фильтр
        self.assertEqual(len(result), 3)


class TestEventFilterEvaluator(unittest.TestCase):
    """Тесты для класса EventFilterEvaluator."""
    
    def setUp(self):
        """Подготовка тестовых данных."""
        self.event = BaseEvent()
        self.event.Severity = 500
        self.event.Message = ua.LocalizedText("Test message")
        self.event.SourceNode = ua.NodeId(1234, 2)
    
    def test_evaluator_matches_true(self):
        """Тест совпадения события с фильтром."""
        evfilter = create_property_filter(
            "Severity",
            ua.FilterOperator.Equals,
            500
        )
        evaluator = EventFilterEvaluator(evfilter)
        self.assertTrue(evaluator.matches(self.event))
    
    def test_evaluator_matches_false(self):
        """Тест несовпадения события с фильтром."""
        evfilter = create_property_filter(
            "Severity",
            ua.FilterOperator.Equals,
            700
        )
        evaluator = EventFilterEvaluator(evfilter)
        self.assertFalse(evaluator.matches(self.event))
    
    def test_evaluator_greater_than(self):
        """Тест оператора GreaterThan."""
        evfilter = create_property_filter(
            "Severity",
            ua.FilterOperator.GreaterThan,
            400
        )
        evaluator = EventFilterEvaluator(evfilter)
        self.assertTrue(evaluator.matches(self.event))
    
    def test_evaluator_less_than(self):
        """Тест оператора LessThan."""
        evfilter = create_property_filter(
            "Severity",
            ua.FilterOperator.LessThan,
            600
        )
        evaluator = EventFilterEvaluator(evfilter)
        self.assertTrue(evaluator.matches(self.event))


class TestEventFilterOperators(unittest.TestCase):
    """Тесты для различных операторов фильтрации."""
    
    def setUp(self):
        """Подготовка тестовых данных."""
        self.event = BaseEvent()
        self.event.Severity = 500
        self.event.Message = ua.LocalizedText("Error in system")
        self.event.SourceNode = ua.NodeId(1234, 2)
    
    def test_in_list_operator(self):
        """Тест оператора InList."""
        evfilter = ua.EventFilter()
        
        # SelectClauses
        op = ua.SimpleAttributeOperand()
        op.AttributeId = ua.AttributeIds.Value
        op.BrowsePath = [ua.QualifiedName("Severity", 0)]
        op.TypeDefinitionId = ua.NodeId(ua.ObjectIds.BaseEventType)
        evfilter.SelectClauses.append(op)
        
        # WhereClause - Severity InList [300, 500, 700]
        cf = ua.ContentFilter()
        el = ua.ContentFilterElement()
        
        # Первый операнд - Severity
        severity_operand = ua.SimpleAttributeOperand()
        severity_operand.BrowsePath = [ua.QualifiedName("Severity", 0)]
        severity_operand.AttributeId = ua.AttributeIds.Value
        severity_operand.TypeDefinitionId = ua.NodeId(ua.ObjectIds.BaseEventType)
        el.FilterOperands.append(ua.ExtensionObject(Body=severity_operand))
        
        # Список значений
        for value in [300, 500, 700]:
            literal = ua.LiteralOperand(Value=ua.Variant(value, ua.VariantType.UInt16))
            el.FilterOperands.append(ua.ExtensionObject(Body=literal))
        
        el.FilterOperator = ua.FilterOperator.InList
        cf.Elements.append(el)
        evfilter.WhereClause = cf
        
        evaluator = EventFilterEvaluator(evfilter)
        self.assertTrue(evaluator.matches(self.event))
    
    def test_between_operator(self):
        """Тест оператора Between."""
        evfilter = ua.EventFilter()
        
        # SelectClauses
        op = ua.SimpleAttributeOperand()
        op.AttributeId = ua.AttributeIds.Value
        op.BrowsePath = [ua.QualifiedName("Severity", 0)]
        op.TypeDefinitionId = ua.NodeId(ua.ObjectIds.BaseEventType)
        evfilter.SelectClauses.append(op)
        
        # WhereClause - Severity Between 400 and 600
        cf = ua.ContentFilter()
        el = ua.ContentFilterElement()
        
        # Первый операнд - Severity
        severity_operand = ua.SimpleAttributeOperand()
        severity_operand.BrowsePath = [ua.QualifiedName("Severity", 0)]
        severity_operand.AttributeId = ua.AttributeIds.Value
        severity_operand.TypeDefinitionId = ua.NodeId(ua.ObjectIds.BaseEventType)
        el.FilterOperands.append(ua.ExtensionObject(Body=severity_operand))
        
        # Нижняя граница
        lower = ua.LiteralOperand(Value=ua.Variant(400, ua.VariantType.UInt16))
        el.FilterOperands.append(ua.ExtensionObject(Body=lower))
        
        # Верхняя граница
        upper = ua.LiteralOperand(Value=ua.Variant(600, ua.VariantType.UInt16))
        el.FilterOperands.append(ua.ExtensionObject(Body=upper))
        
        el.FilterOperator = ua.FilterOperator.Between
        cf.Elements.append(el)
        evfilter.WhereClause = cf
        
        evaluator = EventFilterEvaluator(evfilter)
        self.assertTrue(evaluator.matches(self.event))


class TestEventFilterComplex(unittest.TestCase):
    """Тесты для сложных фильтров."""
    
    def setUp(self):
        """Подготовка тестовых данных."""
        self.event = BaseEvent()
        self.event.Severity = 700
        self.event.Message = ua.LocalizedText("Alarm event")
        self.event.SourceNode = ua.NodeId(1234, 2)
        self.event.EventType = ua.NodeId(ua.ObjectIds.AlarmConditionType)

    def test_root_as_last_element_and(self):
        """
        Фильтр, где корневой элемент (And) находится в последнем элементе массива Elements.
        
        Elements[0]: Severity > 500
        Elements[1]: EventType InList [AlarmConditionType]
        Elements[2]: And(Element(0), Element(1))  <-- корень
        """
        evfilter = ua.EventFilter()

        # SelectClauses (не критично для matches, но добавим для полноты)
        op = ua.SimpleAttributeOperand()
        op.AttributeId = ua.AttributeIds.Value
        op.BrowsePath = [ua.QualifiedName("Severity", 0)]
        op.TypeDefinitionId = ua.NodeId(ua.ObjectIds.BaseEventType)
        evfilter.SelectClauses.append(op)

        cf = ua.ContentFilter()

        # Element 0: Severity > 500
        el0 = ua.ContentFilterElement()
        sev_operand = ua.SimpleAttributeOperand()
        sev_operand.BrowsePath = [ua.QualifiedName("Severity", 0)]
        sev_operand.AttributeId = ua.AttributeIds.Value
        sev_operand.TypeDefinitionId = ua.NodeId(ua.ObjectIds.BaseEventType)
        el0.FilterOperands.append(ua.ExtensionObject(Body=sev_operand))
        el0.FilterOperands.append(
            ua.ExtensionObject(Body=ua.LiteralOperand(Value=ua.Variant(500, ua.VariantType.UInt16)))
        )
        el0.FilterOperator = ua.FilterOperator.GreaterThan
        cf.Elements.append(el0)

        # Element 1: EventType InList [AlarmConditionType]
        el1 = ua.ContentFilterElement()
        etype_operand = ua.SimpleAttributeOperand()
        etype_operand.BrowsePath = [ua.QualifiedName("EventType", 0)]
        etype_operand.AttributeId = ua.AttributeIds.Value
        etype_operand.TypeDefinitionId = ua.NodeId(ua.ObjectIds.BaseEventType)
        el1.FilterOperands.append(ua.ExtensionObject(Body=etype_operand))
        el1.FilterOperands.append(
            ua.ExtensionObject(
                Body=ua.LiteralOperand(
                    Value=ua.Variant(
                        ua.NodeId(ua.ObjectIds.AlarmConditionType),
                        ua.VariantType.NodeId,
                    )
                )
            )
        )
        el1.FilterOperator = ua.FilterOperator.InList
        cf.Elements.append(el1)

        # Element 2: And(ElementOperand(0), ElementOperand(1)) - корень
        el2 = ua.ContentFilterElement()
        el2.FilterOperator = ua.FilterOperator.And
        el2.FilterOperands.append(ua.ExtensionObject(Body=ua.ElementOperand(Index=0)))
        el2.FilterOperands.append(ua.ExtensionObject(Body=ua.ElementOperand(Index=1)))
        cf.Elements.append(el2)

        evfilter.WhereClause = cf

        evaluator = EventFilterEvaluator(evfilter)
        self.assertTrue(evaluator.matches(self.event))


def run_tests():
    """Запуск всех тестов."""
    loader = unittest.TestLoader()
    suite = unittest.TestSuite()
    
    suite.addTests(loader.loadTestsFromTestCase(TestEventFilterBasic))
    suite.addTests(loader.loadTestsFromTestCase(TestEventFilterEvaluator))
    suite.addTests(loader.loadTestsFromTestCase(TestEventFilterOperators))
    suite.addTests(loader.loadTestsFromTestCase(TestEventFilterComplex))
    
    runner = unittest.TextTestRunner(verbosity=2)
    result = runner.run(suite)
    
    return result


if __name__ == "__main__":
    run_tests()

